"""Import logic for blueprint."""
from __future__ import annotations

from dataclasses import dataclass
import html
import re

import voluptuous as vol
import yarl

from homeassistant.core import HomeAssistant
from homeassistant.exceptions import HomeAssistantError
from homeassistant.helpers import aiohttp_client, config_validation as cv
from homeassistant.util import yaml

from .models import Blueprint
from .schemas import is_blueprint_config

COMMUNITY_TOPIC_PATTERN = re.compile(
    r"^https://community.home-assistant.io/t/[a-z0-9-]+/(?P<topic>\d+)(?:/(?P<post>\d+)|)$"
)

COMMUNITY_CODE_BLOCK = re.compile(
    r'<code class="lang-(?P<syntax>[a-z]+)">(?P<content>(?:.|\n)*)</code>', re.MULTILINE
)

GITHUB_FILE_PATTERN = re.compile(
    r"^https://github.com/(?P<repository>.+)/blob/(?P<path>.+)$"
)

COMMUNITY_TOPIC_SCHEMA = vol.Schema(
    {
        "slug": str,
        "title": str,
        "post_stream": {"posts": [{"updated_at": cv.datetime, "cooked": str}]},
    },
    extra=vol.ALLOW_EXTRA,
)


class UnsupportedUrl(HomeAssistantError):
    """When the function doesn't support the url."""


@dataclass(frozen=True)
class ImportedBlueprint:
    """Imported blueprint."""

    suggested_filename: str
    raw_data: str
    blueprint: Blueprint


def _get_github_import_url(url: str) -> str:
    """Convert a GitHub url to the raw content.

    Async friendly.
    """
    if url.startswith("https://raw.githubusercontent.com/"):
        return url

    match = GITHUB_FILE_PATTERN.match(url)

    if match is None:
        raise UnsupportedUrl("Not a GitHub file url")

    repo, path = match.groups()

    return f"https://raw.githubusercontent.com/{repo}/{path}"


def _get_community_post_import_url(url: str) -> str:
    """Convert a forum post url to an import url.

    Async friendly.
    """
    match = COMMUNITY_TOPIC_PATTERN.match(url)
    if match is None:
        raise UnsupportedUrl("Not a topic url")

    _topic, post = match.groups()

    json_url = url

    if post is not None:
        # Chop off post part, ie /2
        json_url = json_url[: -len(post) - 1]

    json_url += ".json"

    return json_url


def _extract_blueprint_from_community_topic(
    url: str,
    topic: dict,
) -> ImportedBlueprint | None:
    """Extract a blueprint from a community post JSON.

    Async friendly.
    """
    block_content = None
    blueprint = None
    post = topic["post_stream"]["posts"][0]

    for match in COMMUNITY_CODE_BLOCK.finditer(post["cooked"]):
        block_syntax, block_content = match.groups()

        if block_syntax not in ("auto", "yaml"):
            continue

        block_content = html.unescape(block_content.strip())

        try:
            data = yaml.parse_yaml(block_content)
        except HomeAssistantError:
            if block_syntax == "yaml":
                raise

            continue

        if not is_blueprint_config(data):
            continue

        blueprint = Blueprint(data)
        break

    if blueprint is None:
        raise HomeAssistantError(
            "No valid blueprint found in the topic. Blueprint syntax blocks need to be marked as YAML or no syntax."
        )

    return ImportedBlueprint(
        f'{post["username"]}/{topic["slug"]}', block_content, blueprint
    )


async def fetch_blueprint_from_community_post(
    hass: HomeAssistant, url: str
) -> ImportedBlueprint | None:
    """Get blueprints from a community post url.

    Method can raise aiohttp client exceptions, vol.Invalid.

    Caller needs to implement own timeout.
    """
    import_url = _get_community_post_import_url(url)
    session = aiohttp_client.async_get_clientsession(hass)

    resp = await session.get(import_url, raise_for_status=True)
    json_resp = await resp.json()
    json_resp = COMMUNITY_TOPIC_SCHEMA(json_resp)
    return _extract_blueprint_from_community_topic(url, json_resp)


async def fetch_blueprint_from_github_url(
    hass: HomeAssistant, url: str
) -> ImportedBlueprint:
    """Get a blueprint from a github url."""
    import_url = _get_github_import_url(url)
    session = aiohttp_client.async_get_clientsession(hass)

    resp = await session.get(import_url, raise_for_status=True)
    raw_yaml = await resp.text()
    data = yaml.parse_yaml(raw_yaml)
    blueprint = Blueprint(data)

    parsed_import_url = yarl.URL(import_url)
    suggested_filename = f"{parsed_import_url.parts[1]}/{parsed_import_url.parts[-1]}"
    if suggested_filename.endswith(".yaml"):
        suggested_filename = suggested_filename[:-5]

    return ImportedBlueprint(suggested_filename, raw_yaml, blueprint)


async def fetch_blueprint_from_github_gist_url(
    hass: HomeAssistant, url: str
) -> ImportedBlueprint:
    """Get a blueprint from a Github Gist."""
    if not url.startswith("https://gist.github.com/"):
        raise UnsupportedUrl("Not a GitHub gist url")

    parsed_url = yarl.URL(url)
    session = aiohttp_client.async_get_clientsession(hass)

    resp = await session.get(
        f"https://api.github.com/gists/{parsed_url.parts[2]}",
        headers={"Accept": "application/vnd.github.v3+json"},
        raise_for_status=True,
    )
    gist = await resp.json()

    blueprint = None
    filename = None
    content = None

    for filename, info in gist["files"].items():
        if not filename.endswith(".yaml"):
            continue

        content = info["content"]
        data = yaml.parse_yaml(content)

        if not is_blueprint_config(data):
            continue

        blueprint = Blueprint(data)
        break

    if blueprint is None:
        raise HomeAssistantError(
            "No valid blueprint found in the gist. The blueprint file needs to end with '.yaml'"
        )

    return ImportedBlueprint(
        f"{gist['owner']['login']}/{filename[:-5]}", content, blueprint
    )


async def fetch_blueprint_from_url(hass: HomeAssistant, url: str) -> ImportedBlueprint:
    """Get a blueprint from a url."""
    for func in (
        fetch_blueprint_from_community_post,
        fetch_blueprint_from_github_url,
        fetch_blueprint_from_github_gist_url,
    ):
        try:
            imported_bp = await func(hass, url)
            imported_bp.blueprint.update_metadata(source_url=url)
            return imported_bp
        except UnsupportedUrl:
            pass

    raise HomeAssistantError("Unsupported url")



"""The Philips TV integration."""
from __future__ import annotations

import asyncio
from datetime import timedelta
import logging
from typing import Any, Callable

from haphilipsjs import ConnectionFailure, PhilipsTV

from homeassistant.components.automation import AutomationActionType
from homeassistant.config_entries import ConfigEntry
from homeassistant.const import (
    CONF_API_VERSION,
    CONF_HOST,
    CONF_PASSWORD,
    CONF_USERNAME,
)
from homeassistant.core import CALLBACK_TYPE, Context, HassJob, HomeAssistant, callback
from homeassistant.helpers.debounce import Debouncer
from homeassistant.helpers.typing import HomeAssistantType
from homeassistant.helpers.update_coordinator import DataUpdateCoordinator

from .const import DOMAIN

PLATFORMS = ["media_player", "remote"]

LOGGER = logging.getLogger(__name__)


async def async_setup(hass: HomeAssistant, config: dict):
    """Set up the Philips TV component."""
    hass.data[DOMAIN] = {}
    return True


async def async_setup_entry(hass: HomeAssistant, entry: ConfigEntry):
    """Set up Philips TV from a config entry."""

    tvapi = PhilipsTV(
        entry.data[CONF_HOST],
        entry.data[CONF_API_VERSION],
        username=entry.data.get(CONF_USERNAME),
        password=entry.data.get(CONF_PASSWORD),
    )

    coordinator = PhilipsTVDataUpdateCoordinator(hass, tvapi)

    await coordinator.async_refresh()
    hass.data[DOMAIN][entry.entry_id] = coordinator

    for platform in PLATFORMS:
        hass.async_create_task(
            hass.config_entries.async_forward_entry_setup(entry, platform)
        )

    return True


async def async_unload_entry(hass: HomeAssistant, entry: ConfigEntry):
    """Unload a config entry."""
    unload_ok = all(
        await asyncio.gather(
            *[
                hass.config_entries.async_forward_entry_unload(entry, platform)
                for platform in PLATFORMS
            ]
        )
    )
    if unload_ok:
        hass.data[DOMAIN].pop(entry.entry_id)

    return unload_ok


class PluggableAction:
    """A pluggable action handler."""

    def __init__(self, update: Callable[[], None]):
        """Initialize."""
        self._update = update
        self._actions: dict[Any, AutomationActionType] = {}

    def __bool__(self):
        """Return if we have something attached."""
        return bool(self._actions)

    @callback
    def async_attach(self, action: AutomationActionType, variables: dict[str, Any]):
        """Attach a device trigger for turn on."""

        @callback
        def _remove():
            del self._actions[_remove]
            self._update()

        job = HassJob(action)

        self._actions[_remove] = (job, variables)
        self._update()

        return _remove

    async def async_run(self, hass: HomeAssistantType, context: Context | None = None):
        """Run all turn on triggers."""
        for job, variables in self._actions.values():
            hass.async_run_hass_job(job, variables, context)


class PhilipsTVDataUpdateCoordinator(DataUpdateCoordinator[None]):
    """Coordinator to update data."""

    def __init__(self, hass, api: PhilipsTV) -> None:
        """Set up the coordinator."""
        self.api = api
        self._notify_future: asyncio.Task | None = None

        @callback
        def _update_listeners():
            for update_callback in self._listeners:
                update_callback()

        self.turn_on = PluggableAction(_update_listeners)

        super().__init__(
            hass,
            LOGGER,
            name=DOMAIN,
            update_interval=timedelta(seconds=30),
            request_refresh_debouncer=Debouncer(
                hass, LOGGER, cooldown=2.0, immediate=False
            ),
        )

    async def _notify_task(self):
        while self.api.on and self.api.notify_change_supported:
            res = await self.api.notifyChange(130)
            if res:
                self.async_set_updated_data(None)
            elif res is None:
                LOGGER.debug("Aborting notify due to unexpected return")
                break

    @callback
    def _async_notify_stop(self):
        if self._notify_future:
            self._notify_future.cancel()
            self._notify_future = None

    @callback
    def _async_notify_schedule(self):
        if (
            (self._notify_future is None or self._notify_future.done())
            and self.api.on
            and self.api.notify_change_supported
        ):
            self._notify_future = asyncio.create_task(self._notify_task())

    @callback
    def async_remove_listener(self, update_callback: CALLBACK_TYPE) -> None:
        """Remove data update."""
        super().async_remove_listener(update_callback)
        if not self._listeners:
            self._async_notify_stop()

    @callback
    def _async_stop_refresh(self, event: asyncio.Event) -> None:
        super()._async_stop_refresh(event)
        self._async_notify_stop()

    @callback
    async def _async_update_data(self):
        """Fetch the latest data from the source."""
        try:
            await self.api.update()
            self._async_notify_schedule()
        except ConnectionFailure:
            pass



""" ExtaLife JSON API wrapper library. Enables device control, discovery and status fetching from EFC-01 controller """

import asyncio
from asyncio.events import AbstractEventLoop
import json
import logging
import socket

import attr

_LOGGER = logging.getLogger(__name__)

# controller info
PRODUCT_MANUFACTURER = "ZAMEL"
PRODUCT_SERIES = "Exta Life"
PRODUCT_SERIES_EXTA_FREE = "Exta Free"
PRODUCT_CONTROLLER_MODEL = "EFC-01"

MODEL_RNK22 = "RNK-22"
MODEL_RNK22_TEMP_SENSOR = "RNK-22 temperature sensor"
MODEL_RNK24 = "RNK-24"
MODEL_RNK24_TEMP_SENSOR = "RNK-24 temperature sensor"
MODEL_P4572 = "P-457/2"
MODEL_P4574 = "P-457/4"
MODEL_P4578 = "P-457/8"
MODEL_P45736 = "P457/36"
MODEL_LEDIX_P260 = "ledix touch control P260"
MODEL_ROP21 = "ROP-21"
MODEL_ROP22 = "ROP-22"
MODEL_SRP22 = "SRP-22"
MODEL_RDP21 = "RDP-21"
MODEL_GKN01 = "GKN-01"
MODEL_ROP27 = "ROP-27"
MODEL_RGT01 = "RGT-01"
MODEL_RNM24 = "RNM-24"
MODEL_RNP21 = "RNP-21"
MODEL_RNP22 = "RNP-22"
MODEL_RCT21 = "RCT-21"
MODEL_RCT22 = "RCT-22"
MODEL_ROG21 = "ROG-21"
MODEL_ROM22 = "ROM-22"
MODEL_ROM24 = "ROM-24"
MODEL_SRM22 = "SRM-22"
MODEL_SLR21 = "SLR-21"
MODEL_SLR22 = "SLR-22"
MODEL_RCM21 = "RCM-21"
MODEL_MEM21 = "MEM-21"
MODEL_RCR21 = "RCR-21"
MODEL_RCZ21 = "RCZ-21"
MODEL_SLM21 = "SLM-21"
MODEL_SLM22 = "SLM-22"
MODEL_RCK21 = "RCK-21"
MODEL_ROB21 = "ROB-21"
MODEL_P501 = "P-501"
MODEL_P520 = "P-520"
MODEL_P521L = "P-521L"
MODEL_BULIK_DRS985 = "bulik DRS-985"

# Exta Free
MODEL_ROP01 = "ROP-01"
MODEL_ROP02 = "ROP-02"
MODEL_ROM01 = "ROM-01"
MODEL_ROM10 = "ROM-10"
MODEL_ROP05 = "ROP-05"
MODEL_ROP06 = "ROP-06"
MODEL_ROP07 = "ROP-07"
MODEL_RWG01 = "RWG-01"
MODEL_ROB01 = "ROB-01"
MODEL_SRP02 = "SRP-02"
MODEL_RDP01 = "RDP-01"
MODEL_RDP02 = "RDP-02"
MODEL_RDP11 = "RDP-11"
MODEL_SRP03 = "SRP-03"

# device types string mapping
DEVICE_MAP_TYPE_TO_MODEL = {
    1: MODEL_RNK22,
    2: MODEL_RNK22_TEMP_SENSOR,
    3: MODEL_RNK24,
    4: MODEL_RNK22_TEMP_SENSOR,
    5: MODEL_P4572,
    6: MODEL_P4574,
    7: MODEL_P4578,
    8: MODEL_P45736,
    9: MODEL_LEDIX_P260,
    10: MODEL_ROP21,
    11: MODEL_ROP22,
    12: MODEL_SRP22,
    13: MODEL_RDP21,
    14: MODEL_GKN01,
    15: MODEL_ROP27,
    16: MODEL_RGT01,
    17: MODEL_RNM24,
    18: MODEL_RNP21,
    19: MODEL_RNP22,
    20: MODEL_RCT21,
    21: MODEL_RCT22,
    22: MODEL_ROG21,
    23: MODEL_ROM22,
    24: MODEL_ROM24,
    25: MODEL_SRM22,
    26: MODEL_SLR21,
    27: MODEL_SLR22,
    28: MODEL_RCM21,
    35: MODEL_MEM21,
    41: MODEL_RCR21,
    42: MODEL_RCZ21,
    45: MODEL_SLM21,
    46: MODEL_SLM22,
    47: MODEL_RCK21,
    48: MODEL_ROB21,
    51: MODEL_P501,
    52: MODEL_P520,
    53: MODEL_P521L,
    238: MODEL_BULIK_DRS985,
    # Exta Free
    326: MODEL_ROP01,
    327: MODEL_ROP02,
    328: MODEL_ROM01,
    329: MODEL_ROM10,
    330: MODEL_ROP05,
    331: MODEL_ROP06,
    332: MODEL_ROP07,
    333: MODEL_RWG01,
    334: MODEL_ROB01,
    335: MODEL_SRP02,
    336: MODEL_RDP01,
    337: MODEL_RDP02,
    338: MODEL_RDP11,
    339: MODEL_SRP03,
}

# reverse lookup
MODEL_MAP_MODEL_TO_TYPE = {v: k for k, v in DEVICE_MAP_TYPE_TO_MODEL.items()}

# device type (channel_data.data.type)
DEVICE_ARR_SENS_TEMP = [2, 4, 20, 21]
DEVICE_ARR_SENS_LIGHT = []
DEVICE_ARR_SENS_HUMID = []
DEVICE_ARR_SENS_PRESSURE = []
DEVICE_ARR_SENS_MULTI = [28]
DEVICE_ARR_SENS_WATER = [42]
DEVICE_ARR_SENS_MOTION = [41]
DEVICE_ARR_SENS_OPENCLOSE = [47]
DEVICE_ARR_SENS_ENERGY_METER = [35]
DEVICE_ARR_SENS_GATE_CONTROLLER = [48]
DEVICE_ARR_SWITCH = [10, 11, 22, 23, 24]
DEVICE_ARR_COVER = [12, 25]
DEVICE_ARR_LIGHT = [13, 26, 45, 27, 46]
DEVICE_ARR_LIGHT_RGB = []  # RGB only
DEVICE_ARR_LIGHT_RGBW = [27, 38]
DEVICE_ARR_LIGHT_EFFECT = [27, 38]
DEVICE_ARR_CLIMATE = [16]
DEVICE_ARR_REPEATER = [237]
DEVICE_ARR_TRANS_REMOTE = [5, 6, 7, 8, 51, 52, 53]
DEVICE_ARR_TRANS_NORMAL_BATTERY = [1, 3, 19]
DEVICE_ARR_TRANS_NORMAL_MAINS = [17, 18]

# Exta Free devices
DEVICE_ARR_EXTA_FREE_RECEIVER = [80]
DEVICE_ARR_EXTA_FREE_SWITCH = [326, 327, 328, 329, 330, 331, 332, 333, 334]
DEVICE_ARR_EXTA_FREE_COVER = [335, 339]
DEVICE_ARR_EXTA_FREE_LIGHT = [336, 337]
DEVICE_ARR_EXTA_FREE_RGB = [338]

DEVICE_ARR_ALL_EXFREE_SWITCH = [*DEVICE_ARR_EXTA_FREE_SWITCH]
DEVICE_ARR_ALL_EXFREE_LIGHT = [*DEVICE_ARR_EXTA_FREE_LIGHT, *DEVICE_ARR_EXTA_FREE_RGB]
DEVICE_ARR_ALL_EXFREE_COVER = [*DEVICE_ARR_EXTA_FREE_COVER]

# union of all subtypes
DEVICE_ARR_ALL_SWITCH = [*DEVICE_ARR_SWITCH, *DEVICE_ARR_ALL_EXFREE_SWITCH]
DEVICE_ARR_ALL_LIGHT = [
    *DEVICE_ARR_LIGHT,
    *DEVICE_ARR_LIGHT_RGB,
    *DEVICE_ARR_LIGHT_RGBW,
    *DEVICE_ARR_ALL_EXFREE_LIGHT,
]
DEVICE_ARR_ALL_COVER = [
    *DEVICE_ARR_COVER,
    *DEVICE_ARR_SENS_GATE_CONTROLLER,
    *DEVICE_ARR_ALL_EXFREE_COVER,
]
DEVICE_ARR_ALL_CLIMATE = [*DEVICE_ARR_CLIMATE]
DEVICE_ARR_ALL_TRANSMITTER = [
    *DEVICE_ARR_TRANS_REMOTE,
    *DEVICE_ARR_TRANS_NORMAL_BATTERY,
    *DEVICE_ARR_TRANS_NORMAL_MAINS,
]
DEVICE_ARR_ALL_IGNORE = [*DEVICE_ARR_REPEATER]

# measurable magnitude/quantity:
DEVICE_ARR_ALL_SENSOR_MEAS = [
    *DEVICE_ARR_SENS_TEMP,
    *DEVICE_ARR_SENS_HUMID,
    *DEVICE_ARR_SENS_ENERGY_METER,
]
# binary sensors:
DEVICE_ARR_ALL_SENSOR_BINARY = [
    *DEVICE_ARR_SENS_WATER,
    *DEVICE_ARR_SENS_MOTION,
    *DEVICE_ARR_SENS_OPENCLOSE,
]
DEVICE_ARR_ALL_SENSOR_MULTI = [*DEVICE_ARR_SENS_MULTI]
DEVICE_ARR_ALL_SENSOR = [
    *DEVICE_ARR_ALL_SENSOR_MEAS,
    *DEVICE_ARR_ALL_SENSOR_BINARY,
    *DEVICE_ARR_ALL_SENSOR_MULTI,
]

# list of device types mapped into `light` platform in HA
DEVICE_ICON_ARR_LIGHT = [
    15,
    13,
    8,
    9,
    14,
    16,
    17,
]  # override device and type rules based on icon; force 'light' device for some icons,


# but only when device was detected preliminarly as switch; 28 =LED


class ExtaLifeAPI:
    """ Main API class: wrapper for communication with controller """

    # Commands
    CMD_LOGIN = 1
    CMD_CONTROL_DEVICE = 20
    CMD_FETCH_RECEIVERS = 37
    CMD_FETCH_SENSORS = 38
    CMD_FETCH_TRANSMITTERS = 39
    CMD_FETCH_NETW_SETTINGS = 102
    CMD_FETCH_EXTAFREE = 203
    CMD_VERSION = 151
    CMD_RESTART = 150

    # Actions
    ACTN_TURN_ON = "TURN_ON"
    ACTN_TURN_OFF = "TURN_OFF"
    ACTN_SET_BRI = "SET_BRIGHTNESS"
    ACTN_SET_RGB = "SET_COLOR"
    ACTN_SET_POS = "SET_POSITION"
    ACTN_SET_TMP = "SET_TEMPERATURE"
    ACTN_STOP = "STOP"
    ACTN_OPEN = "UP"
    ACTN_CLOSE = "DOWN"
    ACTN_SET_SLR_MODE = "SET_MODE"
    ACTN_SET_RGT_MODE_MANUAL = "RGT_SET_MODE_MANUAL"
    ACTN_SET_RGT_MODE_AUTO = "RGT_SET_MODE_AUTO"

    # Exta Free Actions
    ACTN_EXFREE_TURN_ON_PRESS = "TURN_ON_PRESS"
    ACTN_EXFREE_TURN_ON_RELEASE = "TURN_ON_RELEASE"
    ACTN_EXFREE_TURN_OFF_PRESS = "TURN_OFF_PRESS"
    ACTN_EXFREE_TURN_OFF_RELEASE = "TURN_OFF_RELEASE"
    ACTN_EXFREE_UP_PRESS = "UP_PRESS"
    ACTN_EXFREE_UP_RELEASE = "UP_RELEASE"
    ACTN_EXFREE_DOWN_PRESS = "DOWN_PRESS"
    ACTN_EXFREE_DOWN_RELEASE = "DOWN_RELEASE"
    ACTN_EXFREE_BRIGHT_UP_PRESS = "BRIGHT_UP_PRESS"
    ACTN_EXFREE_BRIGHT_UP_RELEASE = "BRIGHT_UP_RELEASE"
    ACTN_EXFREE_BRIGHT_DOWN_PRESS = "BRIGHT_DOWN_PRESS"
    ACTN_EXFREE_BRIGHT_DOWN_RELEASE = "BRIGHT_DOWN_RELEASE"

    # Channel Types
    CHN_TYP_RECEIVERS = "receivers"
    CHN_TYP_SENSORS = "sensors"
    CHN_TYP_TRANSMITTERS = "transmitters"
    CHN_TYP_EXFREE_RECEIVERS = "exta_free_receivers"

    def __init__(
        self,
        loop: AbstractEventLoop,
        on_notification_callback=None,
        on_connect_callback=None,
        on_disconnect_callback=None,
    ):
        """API Object constructor

        on_connect - optional callback for notifications when API connects to the controller and performs successfull login

        on_disconnect - optional callback for notifications when API loses connection to the controller"""

        self.tcp: TCPAdapter = None
        self._mac = None
        self._sw_version: str = None
        self._name: str = None

        # set on_connect callback to notify caller
        self._on_connect_callback = on_connect_callback
        self._on_disconnect_callback = on_disconnect_callback
        self._on_notification_callback = on_notification_callback

        self._is_connected = False

        self._loop: AbstractEventLoop = loop

    async def async_connect(self, user, password, host=None):
        self._host = host
        self._user = user
        self._password = password

        # perform controller autodiscovery if no IP specified
        if self._host is None or self._host == "":
            self._host = await self._loop.run_in_executor(
                None, TCPAdapter.discover_controller
            )

        # check if still None after autodiscovery
        if not self._host:
            raise TCPConnError("Could not find controller IP via autodiscovery")

        ConnectionParams.host = self._host
        ConnectionParams.user = self._user
        ConnectionParams.password = self._password
        ConnectionParams.eventloop = self._loop
        ConnectionParams.keepalive = 8  # ping period; in seconds
        ConnectionParams.on_notification_callback = self._async_on_notification_callback
        ConnectionParams.on_connect_callback = (
            self._async_on_tcp_connect_callback
        )  # self._on_connect_callback
        ConnectionParams.on_disconnect_callback = self._async_on_tcp_disconnect_callback

        # init TCP adapter and try to connect
        self._connection = TCPAdapter(ConnectionParams)

        # connect and login - may raise TCPConnErr
        _LOGGER.debug("Connecting to controller using IP: %s", self._host)
        await self._connection.async_connect()

        resp = await self._connection.async_login()

        # check response if login succeeded
        if resp[0]["status"] != "success":
            raise TCPConnError(resp)

        # determine controller MAC as its unique identifier
        self._mac = await self.async_get_mac()

        return True

    async def async_reconnect(self):
        """ Reconnect with existing connection parameters """
        return await self.async_connect(self._user, self._password, self._host)

    @property
    def host(self):
        return self._host

    async def _async_on_tcp_connect_callback(self):
        """ Called when connectivity is (re)established and logged on successfully """
        self._is_connected = True
        # refresh software version info
        await self.async_get_version_info()
        await self.async_get_name()

        if self._on_connect_callback is not None:
            await self._loop.run_in_executor(None, self._on_connect_callback)

    async def _async_on_tcp_disconnect_callback(self):
        """ Called when connectivity is lost """
        self._is_connected = False

        if self._on_disconnect_callback is not None:
            await self._loop.run_in_executor(None, self._on_disconnect_callback)

    async def _async_on_notification_callback(self, data):
        """ Called when notification from the controller is received """
        if (
            self._on_notification_callback(data) is not None
            and data.get("command") == self.CMD_CONTROL_DEVICE
        ):
            # forward only device status changes to the listener
            self._on_notification_callback(data)

    def set_notification_callback(self, callback):
        """ update Notification callback assignment """
        self._on_notification_callback = callback

    @property
    def is_connected(self) -> bool:
        """ Returns True or False depending of the connection is alive and user is logged on """
        return self._is_connected

    @classmethod
    def discover_controller(cls):
        """ Returns controller IP address if found, otherwise None"""
        return TCPAdapter.discover_controller()

    @property
    def sw_version(self) -> str:
        return self._sw_version

    async def async_get_version_info(self):
        """ Get controller software version """
        cmd_data = {"data": None}
        try:
            resp = await self._connection.async_execute_command(
                self.CMD_VERSION, cmd_data
            )
            self._sw_version = resp[0]["data"]["new_version"]
            return self._sw_version

        except TCPCmdError:
            _LOGGER.error("Command %s could not be executed", self.CMD_VERSION)
            return

    async def async_get_mac(self):
        from getmac import get_mac_address

        # get EFC-01 controller MAC address
        return await self._loop.run_in_executor(None, get_mac_address, None, self._host)

    @property
    def mac(self):
        return self._mac

    async def async_get_network_settings(self):
        """ Executes command 102 to get network settings and controller name """
        try:
            cmd = self.CMD_FETCH_NETW_SETTINGS
            resp = await self._connection.async_execute_command(cmd, None)
            return resp[0].get("data")

        except TCPCmdError:
            _LOGGER.error("Command %s could not be executed", cmd)
            return None

    async def async_get_name(self):
        """ Get controller name """
        data = await self.async_get_network_settings()
        self._name = data.get("name") if data else None
        return self._name

    @property
    def name(self) -> str:
        """ Get controller name from buffer """
        return self._name

    async def async_get_channels(
        self,
        include=(
            CHN_TYP_RECEIVERS,
            CHN_TYP_SENSORS,
            CHN_TYP_TRANSMITTERS,
            CHN_TYP_EXFREE_RECEIVERS,
        ),
    ):
        """
        Get list of dicts of Exta Life channels consisting of native Exta Life TCP JSON
        data, but with transformed data model. Each channel will have native channel info
        AND device info. 2 channels of the same device will have the same device attributes
        """
        try:
            channels = list()
            if self.CHN_TYP_RECEIVERS in include:
                cmd = self.CMD_FETCH_RECEIVERS
                resp = await self._connection.async_execute_command(cmd, None)
                # here is where the magic happens - transform TCP JSON data into API channel representation
                channels.extend(self._get_channels_int(resp))

            if self.CHN_TYP_SENSORS in include:
                cmd = self.CMD_FETCH_SENSORS
                resp = await self._connection.async_execute_command(cmd, None)
                channels.extend(self._get_channels_int(resp))

            if self.CHN_TYP_TRANSMITTERS in include:
                cmd = self.CMD_FETCH_TRANSMITTERS
                resp = await self._connection.async_execute_command(cmd, None)
                channels.extend(self._get_channels_int(resp, dummy_ch=True))

            if self.CHN_TYP_EXFREE_RECEIVERS in include:
                cmd = self.CMD_FETCH_EXTAFREE
                resp = await self._connection.async_execute_command(cmd, None)
                channels.extend(self._get_channels_int(resp))

            return channels

        except TCPCmdError:
            _LOGGER.error("Command %s could not be executed", cmd)
            return None

    @classmethod
    def _get_channels_int(cls, data_js, dummy_ch=False):

        def_channel = None
        if dummy_ch:
            def_channel = "#"
        channels = []  # list of JSON dicts
        for cmd in data_js:
            for device in cmd["data"]["devices"]:
                dev = device.copy()

                if dev.get("exta_free_device"):
                    dev["type"] = (
                        int(dev["state"][0]["exta_free_type"]) + 300
                    )  # do the same as the Exta Life app does - add 300
                    # to move identifiers to Exta Life "namespace"

                dev.pop("state")
                for state in device["state"]:
                    ch_no = (
                        state.get("channel", def_channel)
                        if def_channel
                        else state["channel"]
                    )
                    channel = {
                        # API channel, not TCP channel
                        "id": str(device["id"])
                        + "-"
                        + str(state.get("channel", def_channel)),
                        "data": {**state, **dev},
                    }
                    channels.append(channel)
        return channels

    async def async_execute_action(self, action, channel_id, **fields):
        """Execute action/command in controller
        action - action to be performed. See ACTN_* constants
        channel_id - concatenation of device id and channel number e.g. '1-1'
        **fields - fields of the native JSON command e.g. value, mode, mode_val etc

        Returns array of dicts converted from JSON or None if error occured
        """
        MAP_ACION_STATE = {
            # Exta Life:
            ExtaLifeAPI.ACTN_TURN_ON: 1,
            ExtaLifeAPI.ACTN_TURN_OFF: 0,
            ExtaLifeAPI.ACTN_OPEN: 1,
            ExtaLifeAPI.ACTN_CLOSE: 0,
            ExtaLifeAPI.ACTN_STOP: 2,
            ExtaLifeAPI.ACTN_SET_POS: None,
            ExtaLifeAPI.ACTN_SET_RGT_MODE_AUTO: 0,
            ExtaLifeAPI.ACTN_SET_RGT_MODE_MANUAL: 1,
            ExtaLifeAPI.ACTN_SET_TMP: 1,
            # Exta Free:
            ExtaLifeAPI.ACTN_EXFREE_TURN_ON_PRESS: 1,
            ExtaLifeAPI.ACTN_EXFREE_TURN_ON_RELEASE: 2,
            ExtaLifeAPI.ACTN_EXFREE_TURN_OFF_PRESS: 3,
            ExtaLifeAPI.ACTN_EXFREE_TURN_OFF_RELEASE: 4,
            ExtaLifeAPI.ACTN_EXFREE_UP_PRESS: 1,
            ExtaLifeAPI.ACTN_EXFREE_UP_RELEASE: 2,
            ExtaLifeAPI.ACTN_EXFREE_DOWN_PRESS: 3,
            ExtaLifeAPI.ACTN_EXFREE_DOWN_RELEASE: 4,
            ExtaLifeAPI.ACTN_EXFREE_BRIGHT_UP_PRESS: 1,
            ExtaLifeAPI.ACTN_EXFREE_BRIGHT_UP_RELEASE: 2,
            ExtaLifeAPI.ACTN_EXFREE_BRIGHT_DOWN_PRESS: 3,
            ExtaLifeAPI.ACTN_EXFREE_BRIGHT_DOWN_RELEASE: 4,
        }
        ch_id, channel = channel_id.split("-")
        ch_id = int(ch_id)
        channel = int(channel)

        cmd_data = {
            "id": ch_id,
            "channel": channel,
            "state": MAP_ACION_STATE.get(action),
        }
        # this assumes the right fields are passed to the API
        cmd_data.update(**fields)

        try:
            cmd = self.CMD_CONTROL_DEVICE
            resp = await self._connection.async_execute_command(cmd, cmd_data)

            _LOGGER.debug("JSON response for command %s: %s", cmd, resp)

            return resp
        except TCPCmdError as err:
            # _LOGGER.error("Command %s could not be executed", cmd)
            _LOGGER.exception(err)
            return None

    async def async_restart(self):
        """ Restart EFC-01 """
        try:
            cmd = self.CMD_RESTART
            cmd_data = dict()

            resp = await self._connection.async_execute_command(cmd, cmd_data)

            _LOGGER.debug("JSON response for command %s: %s", cmd, resp)

            return resp
        except TCPCmdError:
            _LOGGER.error("Command %s could not be executed", cmd)
            return None

    async def disconnect(self):
        """ Disconnect from the controller and stop message tasks """
        await self._connection.async_stop(True)

    def get_tcp_adapter(self):
        return self._connection


class TCPConnError(Exception):
    def __init__(self, data=None, previous=None):
        super().__init__()
        self.data = data
        self.error_code = None
        self.previous = previous
        if data:
            data = data[-1].get("data") if isinstance(data[-1], dict) else None
            self.error_code = None if not data else data.get("code")


class TCPCmdError(Exception):
    def __init__(self, data=None):
        super().__init__()
        self.data = data
        self.error_code = None
        if data:
            data = data[-1].get("data") if isinstance(data[-1], dict) else None
            self.error_code = None if not data else data.get("code")


@attr.s
class ConnectionParams:
    eventloop = attr.ib(type=asyncio.events.AbstractEventLoop)
    host = attr.ib(type=str)
    user = attr.ib(type=str)
    password = attr.ib(type=str)
    on_connect_callback = None
    on_disconnect_callback = None
    on_notification_callback = None
    keepalive = attr.ib(type=float)


class APIMessage:
    def __init__(self):
        self.command = ""
        self.data = dict()


class APIRequest(APIMessage):
    def __init__(self, command: str, data: dict):
        super().__init__()
        self.command = command
        self.data = data

    # def from_dict(self, request: dict):
    #     self.command = request.get("command")
    #     self.data = request.get("data")

    def as_dict(self):
        return {"command": self.command, "data": self.data}

    def as_json(self):
        return json.dumps(self.as_dict())


class APIResponse(APIMessage):
    def __init__(self, json: dict):
        super().__init__()
        self._as_dict = json
        self.command = json.get("command")
        self.data = json.get("data")
        self.status = json.get("status")

    @classmethod
    def from_json(cls, json_str: str):
        # print(json_str[:-1])
        json_dict = json.loads(json_str[:-1])

        return APIResponse(json_dict)

    def as_dict(self):
        return self._as_dict


class TCPAdapter:
    TCP_BUFF_SIZE = 8192
    EFC01_PORT = 20400

    _cmd_in_execution = False

    def __init__(self, params: ConnectionParams):

        from datetime import datetime

        self._params = params
        self.user = None
        self.password = None
        self.host = None

        self._on_connect_callback = params.on_connect_callback
        self._on_disconnect_callback = params.on_disconnect_callback

        self.tcp = None

        self._connected = False
        self._stopped = False
        self._authenticated = False
        self._tcp_reader: asyncio.StreamReader = (
            None
        )  # type: Optional[asyncio.StreamReader]
        self._tcp_writer: asyncio.StreamWriter = (
            None
        )  # type: Optional[asyncio.StreamWriter]
        self._write_lock = asyncio.Lock()
        self._cmd_exec_lock = asyncio.Lock()
        self._running_task = None
        self._socket = None
        self._socket_connected = False
        self._ping_task = None

        self._tcp_last_write = datetime.now()

        self._message_handlers = []

    def _start_ping(self) -> None:
        """ Perform "smart" ping task. Send ping if nothing was send to socket in the last keepalive-time period """

        self._ping_task = self._params.eventloop.create_task(self._ping_())

    async def _ping_(self) -> None:
        from datetime import datetime, timedelta

        while self._connected:
            last_write = (datetime.now() - self._tcp_last_write).seconds

            if last_write < self._params.keepalive:
                period = self._params.keepalive - last_write
                await asyncio.sleep(period)
                continue

            if not self._connected:
                break

            try:
                await self.async_ping()
            except TCPConnError:
                _LOGGER.error("%s: Ping Failed!", self._params.address)
                await self._async_on_error()
                break

        _LOGGER.debug("_ping_() - task ends")

    async def async_ping(self) -> None:
        self._check_connected()
        msg = " " + chr(3)
        await self.async_send_message(msg.encode())

    async def _async_write(self, data: bytes) -> None:
        from datetime import datetime

        if not self._socket_connected:
            raise TCPConnError("Socket is not connected")
        try:
            async with self._write_lock:
                self._tcp_writer.write(data)
                self._tcp_last_write = datetime.now()
                await self._tcp_writer.drain()
        except OSError as err:
            await self._async_on_error()
            raise TCPConnError(f"Error while writing data: {err}")

    async def async_send_message(self, msg) -> None:

        _LOGGER.debug("Sending:  %s", str(msg))
        await self._async_write(bytes(msg))

    async def async_send_message_await_response(
        self, send_msg, command: str, timeout: float = 30.0
    ):  # -> Any:
        """ Send message to controller and await response """
        # prevent controller overloading and command loss - wait until finished (lock released)
        async with self._cmd_exec_lock:
            fut = self._params.eventloop.create_future()
            responses = []

            def on_message(resp: APIResponse):
                _LOGGER.debug("on_message(), resp: %s", resp.as_dict())
                if fut.done():
                    return

                if resp.command != command:
                    return

                if resp.status == "searching":
                    responses.append(resp.as_dict())
                elif resp.status in ("success", "failure", "partial"):
                    responses.append(resp.as_dict())
                    fut.set_result(responses)

            self._message_handlers.append(on_message)
            await self.async_send_message(send_msg)

            try:
                await asyncio.wait_for(fut, timeout)

            except asyncio.TimeoutError:
                if self._stopped:
                    raise TCPConnError("Disconnected while waiting for API response!")
                await self._async_on_error()
                raise TCPConnError("Timeout while waiting for API response!")

            try:
                self._message_handlers.remove(on_message)
            except ValueError:
                pass

            return responses
            # return

    async def async_execute_command(self, command: str, data) -> list:

        # request = {"command": command, "data": data}
        # req = self._json_to_tcp(request)
        req = APIRequest(command, data)
        msg = str(req.as_json() + chr(3)).encode()
        response = await self.async_send_message_await_response(msg, command)

        if len(response) == 0:
            raise TCPConnError("No response received from Controller!")

        return response

    async def _async_recv(self) -> bytes:

        try:
            ret = await self._tcp_reader.readuntil(chr(3).encode())
        except (asyncio.IncompleteReadError, OSError, TimeoutError) as err:
            raise TCPConnError(f"Error while receiving data: {err}")

        return ret

    def _check_connected(self) -> None:
        if not self._connected:
            raise TCPConnError("Not connected!")

    async def _close_socket(self) -> None:
        _LOGGER.debug("entering _close_socket()")
        from datetime import datetime

        if not self._socket_connected:
            return
        async with self._write_lock:
            self._tcp_writer.close()
            self._tcp_writer = None
            self._tcp_reader = None
        if self._socket is not None:
            self._socket.close()

        self._socket_connected = False
        self._connected = False
        self._authenticated = False
        _LOGGER.debug("%s: Closed socket", self._params.host)

    async def async_connect(self):
        """
        Connect to EFC-01 via TCP socket
        """
        if self._stopped:
            raise TCPConnError("Connection is closed!")
        if self._connected:
            raise TCPConnError("Already connected!")

        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._socket.setblocking(False)
        self._socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

        _LOGGER.debug("Connecting to %s:%s", self._params.host, self.EFC01_PORT)
        try:
            coro = self._params.eventloop.sock_connect(
                self._socket, (self._params.host, self.EFC01_PORT)
            )
            await asyncio.wait_for(coro, 30.0)
        except OSError as err:
            await self._async_on_error()
            raise TCPConnError(
                f"Error connecting to {self._params.host}: {err}", previous=err
            )
        except asyncio.TimeoutError:
            await self._async_on_error()
            raise TCPConnError(f"Timeout while connecting to {self._params.host}")

        _LOGGER.debug("%s: Opened socket for", self._params.host)
        self._tcp_reader, self._tcp_writer = await asyncio.open_connection(
            sock=self._socket
        )
        self._socket_connected = True
        self._params.eventloop.create_task(self.async_run_forever())

        _LOGGER.debug("Successfully connected ")

        self._connected = True

        self._start_ping()

    async def async_login(self) -> None:
        """
        Try to log on via command: 1
        return json dictionary with result or exception in case of connection or logon
        problem
        """

        self._check_connected()
        if self._authenticated:
            raise TCPConnError("Already logged in!")

        _LOGGER.debug(
            "Logging in...user: %s, password: %s",
            self._params.user,
            self._params.password,
        )
        resp_js = await self.async_execute_command(
            ExtaLifeAPI.CMD_LOGIN,
            {"password": self._params.password, "login": self._params.user},
        )

        if (
            resp_js[0].get("status") == "failure"
            and resp_js[0].get("data").get("code") == -2
        ):
            # pass
            raise TCPConnError("Invalid password!")

        self._authenticated = True

        _LOGGER.debug("Authenticated")

        await self._async_event_connect()

        return resp_js

    async def async_run_forever(self) -> None:
        while True:
            try:
                await self._async_run_once()
            except TCPConnError as err:
                _LOGGER.info("Error while reading incoming messages: %s", err.data)
                await self._async_on_error()
                break
            except Exception as err:  # pylint: disable=broad-except
                _LOGGER.info(
                    "Unexpected error while reading incoming messages: %s", err
                )
                await self._async_on_error()
                break

        _LOGGER.debug("async_run_forever() - task ends")

    async def _async_run_once(self) -> None:

        raw_msg = await self._async_recv()

        msg = raw_msg.decode()

        resp = APIResponse.from_json(msg)
        _LOGGER.debug("_async_run_once, msg: %s", msg)

        for msg_handler in self._message_handlers[:]:
            msg_handler(resp)

        await self._handle_notification(resp)

    async def _handle_notification(self, resp: APIResponse):
        _LOGGER.debug("_handle_notification(), resp: %s", resp.as_dict())

        # pass only status change notifications to registered listeners
        if (
            resp.status == "notification"
            and self._params.on_notification_callback is not None
        ):
            await self._params.on_notification_callback(resp.as_dict())

    async def _async_on_error(self) -> None:
        await self.async_stop(force=True)

    async def async_stop(self, force: bool = False) -> None:
        _LOGGER.debug("async_stop() self._stopped: %s", self._stopped)
        if self._stopped:
            return

        self._stopped = True
        if self._running_task is not None:
            self._running_task.cancel()

        if self._ping_task is not None:
            self._ping_task.cancel()
            try:
                await self._ping_task
            except asyncio.CancelledError:
                pass

        await self._close_socket()

        await self._async_event_disconnect()

    async def _async_event_connect(self):
        """ Notify of (re)connection by calling provided callback """
        if self._on_connect_callback is not None:
            await self._on_connect_callback()

    async def _async_event_disconnect(self):
        """ Notify of lost connection by calling provided callback """
        if self._on_disconnect_callback is not None:
            await self._on_disconnect_callback()

    @staticmethod
    def discover_controller():
        """
        Perform controller autodiscovery using UDP query
        return IP as string or false if not found
        """
        MCAST_GRP = "225.0.0.1"
        MCAST_PORT = 20401
        import struct

        # sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        server_address = ("", MCAST_PORT)

        # Create the socket
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        # Bind to the server address
        try:
            sock.bind(server_address)
        except OSError as e:
            sock.close()
            sock = None
            _LOGGER.error(
                "Could not connect to receive UDP multicast from EFC-01 on port %s",
                MCAST_PORT,
            )
            return False
        # Tell the operating system to add the socket to the multicast group
        # on all interfaces (join multicast group)
        group = socket.inet_aton(MCAST_GRP)
        mreq = struct.pack("4sL", group, socket.INADDR_ANY)
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)

        sock.settimeout(3)
        try:
            data, address = sock.recvfrom(1024)
        except Exception:
            sock.close()
            return
        sock.close()
        _LOGGER.debug("Got multicast response from EFC-01: %s", str(data.decode()))
        if data == b'{"status":"broadcast","command":0,"data":null}\x03':
            return address[0]  # return IP - array[0]; array[1] is sender's port
        return



"""Ruckus Unleashed DataUpdateCoordinator."""
from datetime import timedelta
import logging

from pyruckus import Ruckus
from pyruckus.exceptions import AuthenticationError

from homeassistant.core import HomeAssistant
from homeassistant.helpers.update_coordinator import DataUpdateCoordinator, UpdateFailed

from .const import (
    API_CLIENTS,
    API_CURRENT_ACTIVE_CLIENTS,
    API_MAC,
    DOMAIN,
    SCAN_INTERVAL,
)

_LOGGER = logging.getLogger(__package__)


class RuckusUnleashedDataUpdateCoordinator(DataUpdateCoordinator):
    """Coordinator to manage data from Ruckus Unleashed client."""

    def __init__(self, hass: HomeAssistant, *, ruckus: Ruckus):
        """Initialize global Ruckus Unleashed data updater."""
        self.ruckus = ruckus

        update_interval = timedelta(seconds=SCAN_INTERVAL)

        super().__init__(
            hass,
            _LOGGER,
            name=DOMAIN,
            update_interval=update_interval,
        )

    async def _fetch_clients(self) -> dict:
        """Fetch clients from the API and format them."""
        clients = await self.hass.async_add_executor_job(
            self.ruckus.current_active_clients
        )
        return {e[API_MAC]: e for e in clients[API_CURRENT_ACTIVE_CLIENTS][API_CLIENTS]}

    async def _async_update_data(self) -> dict:
        """Fetch Ruckus Unleashed data."""
        try:
            return {API_CLIENTS: await self._fetch_clients()}
        except (AuthenticationError, ConnectionError) as error:
            raise UpdateFailed(error) from error



"""Config flow for Home Assistant Supervisor integration."""
import logging

from homeassistant import config_entries

from . import DOMAIN

_LOGGER = logging.getLogger(__name__)


class ConfigFlow(config_entries.ConfigFlow, domain=DOMAIN):
    """Handle a config flow for Home Assistant Supervisor."""

    VERSION = 1
    CONNECTION_CLASS = config_entries.CONN_CLASS_LOCAL_PUSH

    async def async_step_system(self, user_input=None):
        """Handle the initial step."""
        # We only need one Hass.io config entry
        await self.async_set_unique_id(DOMAIN)
        self._abort_if_unique_id_configured()
        return self.async_create_entry(title="Supervisor", data={})



"""Config flow for Gogogate2."""
import dataclasses
import re

from gogogate2_api.common import AbstractInfoResponse, ApiError
from gogogate2_api.const import GogoGate2ApiErrorCode, ISmartGateApiErrorCode
import voluptuous as vol

from homeassistant import config_entries
from homeassistant.config_entries import SOURCE_IMPORT, ConfigFlow
from homeassistant.const import (
    CONF_DEVICE,
    CONF_IP_ADDRESS,
    CONF_PASSWORD,
    CONF_USERNAME,
)

from .common import get_api
from .const import DEVICE_TYPE_GOGOGATE2, DEVICE_TYPE_ISMARTGATE, DOMAIN


class Gogogate2FlowHandler(ConfigFlow, domain=DOMAIN):
    """Gogogate2 config flow."""

    VERSION = 1
    CONNECTION_CLASS = config_entries.CONN_CLASS_LOCAL_POLL

    def __init__(self):
        """Initialize the config flow."""
        self._ip_address = None
        self._device_type = None

    async def async_step_import(self, config_data: dict = None):
        """Handle importing of configuration."""
        result = await self.async_step_user(config_data)
        self._abort_if_unique_id_configured()
        return result

    async def async_step_homekit(self, discovery_info):
        """Handle homekit discovery."""
        await self.async_set_unique_id(discovery_info["properties"]["id"])
        self._abort_if_unique_id_configured({CONF_IP_ADDRESS: discovery_info["host"]})

        ip_address = discovery_info["host"]

        for entry in self._async_current_entries():
            if entry.data.get(CONF_IP_ADDRESS) == ip_address:
                return self.async_abort(reason="already_configured")

        self._ip_address = ip_address
        self._device_type = DEVICE_TYPE_ISMARTGATE
        return await self.async_step_user()

    async def async_step_user(self, user_input: dict = None):
        """Handle user initiated flow."""
        user_input = user_input or {}
        errors = {}

        if user_input:
            api = get_api(user_input)
            try:
                data: AbstractInfoResponse = await api.async_info()
                data_dict = dataclasses.asdict(data)
                title = data_dict.get(
                    "gogogatename", data_dict.get("ismartgatename", "Cover")
                )
                await self.async_set_unique_id(re.sub("\\..*$", "", data.remoteaccess))
                return self.async_create_entry(title=title, data=user_input)

            except ApiError as api_error:
                device_type = user_input[CONF_DEVICE]
                is_invalid_auth = (
                    device_type == DEVICE_TYPE_GOGOGATE2
                    and api_error.code
                    in (
                        GogoGate2ApiErrorCode.CREDENTIALS_NOT_SET,
                        GogoGate2ApiErrorCode.CREDENTIALS_INCORRECT,
                    )
                ) or (
                    device_type == DEVICE_TYPE_ISMARTGATE
                    and api_error.code
                    in (
                        ISmartGateApiErrorCode.CREDENTIALS_NOT_SET,
                        ISmartGateApiErrorCode.CREDENTIALS_INCORRECT,
                    )
                )

                if is_invalid_auth:
                    errors["base"] = "invalid_auth"
                else:
                    errors["base"] = "cannot_connect"

            except Exception:  # pylint: disable=broad-except
                errors["base"] = "cannot_connect"

        if errors and self.source == SOURCE_IMPORT:
            return self.async_abort(reason="cannot_connect")

        return self.async_show_form(
            step_id="user",
            data_schema=vol.Schema(
                {
                    vol.Required(
                        CONF_DEVICE,
                        default=self._device_type
                        or user_input.get(CONF_DEVICE, DEVICE_TYPE_GOGOGATE2),
                    ): vol.In((DEVICE_TYPE_GOGOGATE2, DEVICE_TYPE_ISMARTGATE)),
                    vol.Required(
                        CONF_IP_ADDRESS,
                        default=user_input.get(CONF_IP_ADDRESS, self._ip_address),
                    ): str,
                    vol.Required(
                        CONF_USERNAME, default=user_input.get(CONF_USERNAME, "")
                    ): str,
                    vol.Required(
                        CONF_PASSWORD, default=user_input.get(CONF_PASSWORD, "")
                    ): str,
                }
            ),
            errors=errors,
        )



"""Support for Honeywell Lyric climate platform."""
from __future__ import annotations

import logging
from time import localtime, strftime, time

from aiolyric.objects.device import LyricDevice
from aiolyric.objects.location import LyricLocation
import voluptuous as vol

from homeassistant.components.climate import ClimateEntity
from homeassistant.components.climate.const import (
    ATTR_TARGET_TEMP_HIGH,
    ATTR_TARGET_TEMP_LOW,
    CURRENT_HVAC_COOL,
    CURRENT_HVAC_HEAT,
    CURRENT_HVAC_IDLE,
    CURRENT_HVAC_OFF,
    HVAC_MODE_COOL,
    HVAC_MODE_HEAT,
    HVAC_MODE_HEAT_COOL,
    HVAC_MODE_OFF,
    SUPPORT_PRESET_MODE,
    SUPPORT_TARGET_TEMPERATURE,
)
from homeassistant.config_entries import ConfigEntry
from homeassistant.const import ATTR_TEMPERATURE
from homeassistant.exceptions import HomeAssistantError
from homeassistant.helpers import entity_platform
import homeassistant.helpers.config_validation as cv
from homeassistant.helpers.typing import HomeAssistantType
from homeassistant.helpers.update_coordinator import DataUpdateCoordinator

from . import LyricDeviceEntity
from .const import (
    DOMAIN,
    LYRIC_EXCEPTIONS,
    PRESET_HOLD_UNTIL,
    PRESET_NO_HOLD,
    PRESET_PERMANENT_HOLD,
    PRESET_TEMPORARY_HOLD,
    PRESET_VACATION_HOLD,
)

_LOGGER = logging.getLogger(__name__)

SUPPORT_FLAGS = SUPPORT_TARGET_TEMPERATURE | SUPPORT_PRESET_MODE

LYRIC_HVAC_ACTION_OFF = "EquipmentOff"
LYRIC_HVAC_ACTION_HEAT = "Heat"
LYRIC_HVAC_ACTION_COOL = "Cool"

LYRIC_HVAC_MODE_OFF = "Off"
LYRIC_HVAC_MODE_HEAT = "Heat"
LYRIC_HVAC_MODE_COOL = "Cool"
LYRIC_HVAC_MODE_HEAT_COOL = "Auto"

LYRIC_HVAC_MODES = {
    HVAC_MODE_OFF: LYRIC_HVAC_MODE_OFF,
    HVAC_MODE_HEAT: LYRIC_HVAC_MODE_HEAT,
    HVAC_MODE_COOL: LYRIC_HVAC_MODE_COOL,
    HVAC_MODE_HEAT_COOL: LYRIC_HVAC_MODE_HEAT_COOL,
}

HVAC_MODES = {
    LYRIC_HVAC_MODE_OFF: HVAC_MODE_OFF,
    LYRIC_HVAC_MODE_HEAT: HVAC_MODE_HEAT,
    LYRIC_HVAC_MODE_COOL: HVAC_MODE_COOL,
    LYRIC_HVAC_MODE_HEAT_COOL: HVAC_MODE_HEAT_COOL,
}

HVAC_ACTIONS = {
    LYRIC_HVAC_ACTION_OFF: CURRENT_HVAC_OFF,
    LYRIC_HVAC_ACTION_HEAT: CURRENT_HVAC_HEAT,
    LYRIC_HVAC_ACTION_COOL: CURRENT_HVAC_COOL,
}

SERVICE_HOLD_TIME = "set_hold_time"
ATTR_TIME_PERIOD = "time_period"

SCHEMA_HOLD_TIME = {
    vol.Required(ATTR_TIME_PERIOD, default="01:00:00"): vol.All(
        cv.time_period,
        cv.positive_timedelta,
        lambda td: strftime("%H:%M:%S", localtime(time() + td.total_seconds())),
    )
}


async def async_setup_entry(
    hass: HomeAssistantType, entry: ConfigEntry, async_add_entities
) -> None:
    """Set up the Honeywell Lyric climate platform based on a config entry."""
    coordinator: DataUpdateCoordinator = hass.data[DOMAIN][entry.entry_id]

    entities = []

    for location in coordinator.data.locations:
        for device in location.devices:
            entities.append(
                LyricClimate(
                    coordinator, location, device, hass.config.units.temperature_unit
                )
            )

    async_add_entities(entities, True)

    platform = entity_platform.current_platform.get()

    platform.async_register_entity_service(
        SERVICE_HOLD_TIME,
        SCHEMA_HOLD_TIME,
        "async_set_hold_time",
    )


class LyricClimate(LyricDeviceEntity, ClimateEntity):
    """Defines a Honeywell Lyric climate entity."""

    def __init__(
        self,
        coordinator: DataUpdateCoordinator,
        location: LyricLocation,
        device: LyricDevice,
        temperature_unit: str,
    ) -> None:
        """Initialize Honeywell Lyric climate entity."""
        self._temperature_unit = temperature_unit

        # Setup supported hvac modes
        self._hvac_modes = [HVAC_MODE_OFF]

        # Add supported lyric thermostat features
        if LYRIC_HVAC_MODE_HEAT in device.allowedModes:
            self._hvac_modes.append(HVAC_MODE_HEAT)

        if LYRIC_HVAC_MODE_COOL in device.allowedModes:
            self._hvac_modes.append(HVAC_MODE_COOL)

        if (
            LYRIC_HVAC_MODE_HEAT in device.allowedModes
            and LYRIC_HVAC_MODE_COOL in device.allowedModes
        ):
            self._hvac_modes.append(HVAC_MODE_HEAT_COOL)

        super().__init__(
            coordinator,
            location,
            device,
            f"{device.macID}_thermostat",
            device.name,
            None,
        )

    @property
    def supported_features(self) -> int:
        """Return the list of supported features."""
        return SUPPORT_FLAGS

    @property
    def temperature_unit(self) -> str:
        """Return the unit of measurement."""
        return self._temperature_unit

    @property
    def current_temperature(self) -> float | None:
        """Return the current temperature."""
        return self.device.indoorTemperature

    @property
    def hvac_action(self) -> str:
        """Return the current hvac action."""
        action = HVAC_ACTIONS.get(self.device.operationStatus.mode, None)
        if action == CURRENT_HVAC_OFF and self.hvac_mode != HVAC_MODE_OFF:
            action = CURRENT_HVAC_IDLE
        return action

    @property
    def hvac_mode(self) -> str:
        """Return the hvac mode."""
        return HVAC_MODES[self.device.changeableValues.mode]

    @property
    def hvac_modes(self) -> list[str]:
        """List of available hvac modes."""
        return self._hvac_modes

    @property
    def target_temperature(self) -> float | None:
        """Return the temperature we try to reach."""
        device = self.device
        if not device.hasDualSetpointStatus:
            return device.changeableValues.heatSetpoint
        return None

    @property
    def target_temperature_low(self) -> float | None:
        """Return the upper bound temperature we try to reach."""
        device = self.device
        if device.hasDualSetpointStatus:
            return device.changeableValues.coolSetpoint
        return None

    @property
    def target_temperature_high(self) -> float | None:
        """Return the upper bound temperature we try to reach."""
        device = self.device
        if device.hasDualSetpointStatus:
            return device.changeableValues.heatSetpoint
        return None

    @property
    def preset_mode(self) -> str | None:
        """Return current preset mode."""
        return self.device.changeableValues.thermostatSetpointStatus

    @property
    def preset_modes(self) -> list[str] | None:
        """Return preset modes."""
        return [
            PRESET_NO_HOLD,
            PRESET_HOLD_UNTIL,
            PRESET_PERMANENT_HOLD,
            PRESET_TEMPORARY_HOLD,
            PRESET_VACATION_HOLD,
        ]

    @property
    def min_temp(self) -> float:
        """Identify min_temp in Lyric API or defaults if not available."""
        device = self.device
        if LYRIC_HVAC_MODE_COOL in device.allowedModes:
            return device.minCoolSetpoint
        return device.minHeatSetpoint

    @property
    def max_temp(self) -> float:
        """Identify max_temp in Lyric API or defaults if not available."""
        device = self.device
        if LYRIC_HVAC_MODE_HEAT in device.allowedModes:
            return device.maxHeatSetpoint
        return device.maxCoolSetpoint

    async def async_set_temperature(self, **kwargs) -> None:
        """Set new target temperature."""
        target_temp_low = kwargs.get(ATTR_TARGET_TEMP_LOW)
        target_temp_high = kwargs.get(ATTR_TARGET_TEMP_HIGH)

        device = self.device
        if device.hasDualSetpointStatus:
            if target_temp_low is not None and target_temp_high is not None:
                temp = (target_temp_low, target_temp_high)
            else:
                raise HomeAssistantError(
                    "Could not find target_temp_low and/or target_temp_high in arguments"
                )
        else:
            temp = kwargs.get(ATTR_TEMPERATURE)
        _LOGGER.debug("Set temperature: %s", temp)
        try:
            await self._update_thermostat(self.location, device, heatSetpoint=temp)
        except LYRIC_EXCEPTIONS as exception:
            _LOGGER.error(exception)
        await self.coordinator.async_refresh()

    async def async_set_hvac_mode(self, hvac_mode: str) -> None:
        """Set hvac mode."""
        _LOGGER.debug("Set hvac mode: %s", hvac_mode)
        try:
            await self._update_thermostat(
                self.location, self.device, mode=LYRIC_HVAC_MODES[hvac_mode]
            )
        except LYRIC_EXCEPTIONS as exception:
            _LOGGER.error(exception)
        await self.coordinator.async_refresh()

    async def async_set_preset_mode(self, preset_mode: str) -> None:
        """Set preset (PermanentHold, HoldUntil, NoHold, VacationHold) mode."""
        _LOGGER.debug("Set preset mode: %s", preset_mode)
        try:
            await self._update_thermostat(
                self.location, self.device, thermostatSetpointStatus=preset_mode
            )
        except LYRIC_EXCEPTIONS as exception:
            _LOGGER.error(exception)
        await self.coordinator.async_refresh()

    async def async_set_hold_time(self, time_period: str) -> None:
        """Set the time to hold until."""
        _LOGGER.debug("set_hold_time: %s", time_period)
        try:
            await self._update_thermostat(
                self.location,
                self.device,
                thermostatSetpointStatus=PRESET_HOLD_UNTIL,
                nextPeriodTime=time_period,
            )
        except LYRIC_EXCEPTIONS as exception:
            _LOGGER.error(exception)
        await self.coordinator.async_refresh()



"""Support for MAX! binary sensors via MAX! Cube."""
from homeassistant.components.binary_sensor import (
    DEVICE_CLASS_WINDOW,
    BinarySensorEntity,
)

from . import DATA_KEY


def setup_platform(hass, config, add_entities, discovery_info=None):
    """Iterate through all MAX! Devices and add window shutters."""
    devices = []
    for handler in hass.data[DATA_KEY].values():
        for device in handler.cube.devices:
            # Only add Window Shutters
            if device.is_windowshutter():
                devices.append(MaxCubeShutter(handler, device))

    if devices:
        add_entities(devices)


class MaxCubeShutter(BinarySensorEntity):
    """Representation of a MAX! Cube Binary Sensor device."""

    def __init__(self, handler, device):
        """Initialize MAX! Cube BinarySensorEntity."""
        room = handler.cube.room_by_id(device.room_id)
        self._name = f"{room.name} {device.name}"
        self._cubehandle = handler
        self._device = device

    @property
    def name(self):
        """Return the name of the BinarySensorEntity."""
        return self._name

    @property
    def device_class(self):
        """Return the class of this sensor."""
        return DEVICE_CLASS_WINDOW

    @property
    def is_on(self):
        """Return true if the binary sensor is on/open."""
        return self._device.is_open

    def update(self):
        """Get latest data from MAX! Cube."""
        self._cubehandle.update()



"""Signal handling related helpers."""
import logging
import signal
import sys
from types import FrameType

from homeassistant.const import RESTART_EXIT_CODE
from homeassistant.core import HomeAssistant, callback
from homeassistant.loader import bind_hass

_LOGGER = logging.getLogger(__name__)


@callback
@bind_hass
def async_register_signal_handling(hass: HomeAssistant) -> None:
    """Register system signal handler for core."""
    if sys.platform != "win32":

        @callback
        def async_signal_handle(exit_code: int) -> None:
            """Wrap signal handling.

            * queue call to shutdown task
            * re-instate default handler
            """
            if exit_code == 2:
                # the signal SIGINT - Interrupt, need to be ignored on Android
                print("SIGINT")
                # quick stop access to files - to prevent
                # ProcessKiller: Process xxx (10754) has open file /mnt/media_rw/...
                # ProcessKiller: Sending Interrupt to process 10754
                from homeassistant.components.ais_dom import ais_global

                # 1. quick stop logs
                if ais_global.G_LOG_SETTINGS_INFO is not None:
                    # just to be sure
                    print("logging.shutdown")
                    logging.shutdown()
                    log = logging.getLogger()  # root logger
                    log.handlers.clear()
                    log.disabled = True
                    for hdlr in log.handlers[:]:  # remove all old handlers
                        hdlr.flush()
                        hdlr.close()
                        log.removeHandler(hdlr)
                    print("ais_stop_logs_event")
                    hass.bus.async_fire("ais_stop_logs_event")
                # 2. stop recorder if the db is in file
                if ais_global.G_DB_SETTINGS_INFO is not None:
                    db_url = ais_global.G_DB_SETTINGS_INFO["dbUrl"]
                    if db_url.startswith("sqlite://///data"):
                        print("ais_stop_recorder_event")
                        hass.bus.async_fire("ais_stop_recorder_event")
                # hass.loop.remove_signal_handler(signal.SIGINT)
            else:
                hass.loop.remove_signal_handler(signal.SIGTERM)
                hass.async_create_task(hass.async_stop(exit_code))

        try:
            hass.loop.add_signal_handler(signal.SIGTERM, async_signal_handle, 0)
        except ValueError:
            _LOGGER.warning("Could not bind to SIGTERM")

        try:
            hass.loop.add_signal_handler(signal.SIGINT, async_signal_handle, 2)
        except ValueError:
            _LOGGER.warning("Could not bind to SIGINT")

        try:
            hass.loop.add_signal_handler(
                signal.SIGHUP, async_signal_handle, RESTART_EXIT_CODE
            )
        except ValueError:
            _LOGGER.warning("Could not bind to SIGHUP")

    else:
        old_sigterm = None
        old_sigint = None

        @callback
        def async_signal_handle(exit_code: int, frame: FrameType) -> None:
            """Wrap signal handling.

            * queue call to shutdown task
            * re-instate default handler
            """
            signal.signal(signal.SIGTERM, old_sigterm)
            signal.signal(signal.SIGINT, old_sigint)
            hass.async_create_task(hass.async_stop(exit_code))

        try:
            old_sigterm = signal.signal(signal.SIGTERM, async_signal_handle)
        except ValueError:
            _LOGGER.warning("Could not bind to SIGTERM")

        try:
            old_sigint = signal.signal(signal.SIGINT, async_signal_handle)
        except ValueError:
            _LOGGER.warning("Could not bind to SIGINT")



"""Support for GitHub."""
from datetime import timedelta
import logging

import github
import voluptuous as vol

from homeassistant.components.sensor import PLATFORM_SCHEMA, SensorEntity
from homeassistant.const import (
    ATTR_NAME,
    CONF_ACCESS_TOKEN,
    CONF_NAME,
    CONF_PATH,
    CONF_URL,
)
import homeassistant.helpers.config_validation as cv

_LOGGER = logging.getLogger(__name__)

CONF_REPOS = "repositories"

ATTR_LATEST_COMMIT_MESSAGE = "latest_commit_message"
ATTR_LATEST_COMMIT_SHA = "latest_commit_sha"
ATTR_LATEST_RELEASE_TAG = "latest_release_tag"
ATTR_LATEST_RELEASE_URL = "latest_release_url"
ATTR_LATEST_OPEN_ISSUE_URL = "latest_open_issue_url"
ATTR_OPEN_ISSUES = "open_issues"
ATTR_LATEST_OPEN_PULL_REQUEST_URL = "latest_open_pull_request_url"
ATTR_OPEN_PULL_REQUESTS = "open_pull_requests"
ATTR_PATH = "path"
ATTR_STARGAZERS = "stargazers"
ATTR_FORKS = "forks"
ATTR_CLONES = "clones"
ATTR_CLONES_UNIQUE = "clones_unique"
ATTR_VIEWS = "views"
ATTR_VIEWS_UNIQUE = "views_unique"

DEFAULT_NAME = "GitHub"

SCAN_INTERVAL = timedelta(seconds=300)

REPO_SCHEMA = vol.Schema(
    {vol.Required(CONF_PATH): cv.string, vol.Optional(CONF_NAME): cv.string}
)

PLATFORM_SCHEMA = PLATFORM_SCHEMA.extend(
    {
        vol.Required(CONF_ACCESS_TOKEN): cv.string,
        vol.Optional(CONF_URL): cv.url,
        vol.Required(CONF_REPOS): vol.All(cv.ensure_list, [REPO_SCHEMA]),
    }
)


def setup_platform(hass, config, add_entities, discovery_info=None):
    """Set up the GitHub sensor platform."""
    sensors = []
    for repository in config[CONF_REPOS]:
        data = GitHubData(
            repository=repository,
            access_token=config.get(CONF_ACCESS_TOKEN),
            server_url=config.get(CONF_URL),
        )
        if data.setup_error is True:
            _LOGGER.error(
                "Error setting up GitHub platform. %s",
                "Check previous errors for details",
            )
        else:
            sensors.append(GitHubSensor(data))
    add_entities(sensors, True)


class GitHubSensor(SensorEntity):
    """Representation of a GitHub sensor."""

    def __init__(self, github_data):
        """Initialize the GitHub sensor."""
        self._unique_id = github_data.repository_path
        self._name = None
        self._state = None
        self._available = False
        self._repository_path = None
        self._latest_commit_message = None
        self._latest_commit_sha = None
        self._latest_release_tag = None
        self._latest_release_url = None
        self._open_issue_count = None
        self._latest_open_issue_url = None
        self._pull_request_count = None
        self._latest_open_pr_url = None
        self._stargazers = None
        self._forks = None
        self._clones = None
        self._clones_unique = None
        self._views = None
        self._views_unique = None
        self._github_data = github_data

    @property
    def name(self):
        """Return the name of the sensor."""
        return self._name

    @property
    def unique_id(self):
        """Return unique ID for the sensor."""
        return self._unique_id

    @property
    def state(self):
        """Return the state of the sensor."""
        return self._state

    @property
    def available(self):
        """Return True if entity is available."""
        return self._available

    @property
    def extra_state_attributes(self):
        """Return the state attributes."""
        attrs = {
            ATTR_PATH: self._repository_path,
            ATTR_NAME: self._name,
            ATTR_LATEST_COMMIT_MESSAGE: self._latest_commit_message,
            ATTR_LATEST_COMMIT_SHA: self._latest_commit_sha,
            ATTR_LATEST_RELEASE_URL: self._latest_release_url,
            ATTR_LATEST_OPEN_ISSUE_URL: self._latest_open_issue_url,
            ATTR_OPEN_ISSUES: self._open_issue_count,
            ATTR_LATEST_OPEN_PULL_REQUEST_URL: self._latest_open_pr_url,
            ATTR_OPEN_PULL_REQUESTS: self._pull_request_count,
            ATTR_STARGAZERS: self._stargazers,
            ATTR_FORKS: self._forks,
        }
        if self._latest_release_tag is not None:
            attrs[ATTR_LATEST_RELEASE_TAG] = self._latest_release_tag
        if self._clones is not None:
            attrs[ATTR_CLONES] = self._clones
        if self._clones_unique is not None:
            attrs[ATTR_CLONES_UNIQUE] = self._clones_unique
        if self._views is not None:
            attrs[ATTR_VIEWS] = self._views
        if self._views_unique is not None:
            attrs[ATTR_VIEWS_UNIQUE] = self._views_unique
        return attrs

    @property
    def icon(self):
        """Return the icon to use in the frontend."""
        return "mdi:github"

    def update(self):
        """Collect updated data from GitHub API."""
        self._github_data.update()

        self._name = self._github_data.name
        self._repository_path = self._github_data.repository_path
        self._available = self._github_data.available
        self._latest_commit_message = self._github_data.latest_commit_message
        self._latest_commit_sha = self._github_data.latest_commit_sha
        if self._github_data.latest_release_url is not None:
            self._latest_release_tag = self._github_data.latest_release_url.split(
                "tag/"
            )[1]
        else:
            self._latest_release_tag = None
        self._latest_release_url = self._github_data.latest_release_url
        self._state = self._github_data.latest_commit_sha[0:7]
        self._open_issue_count = self._github_data.open_issue_count
        self._latest_open_issue_url = self._github_data.latest_open_issue_url
        self._pull_request_count = self._github_data.pull_request_count
        self._latest_open_pr_url = self._github_data.latest_open_pr_url
        self._stargazers = self._github_data.stargazers
        self._forks = self._github_data.forks
        self._clones = self._github_data.clones
        self._clones_unique = self._github_data.clones_unique
        self._views = self._github_data.views
        self._views_unique = self._github_data.views_unique


class GitHubData:
    """GitHub Data object."""

    def __init__(self, repository, access_token=None, server_url=None):
        """Set up GitHub."""
        self._github = github

        self.setup_error = False

        try:
            if server_url is not None:
                server_url += "/api/v3"
                self._github_obj = github.Github(access_token, base_url=server_url)
            else:
                self._github_obj = github.Github(access_token)

            self.repository_path = repository[CONF_PATH]

            repo = self._github_obj.get_repo(self.repository_path)
        except self._github.GithubException as err:
            _LOGGER.error("GitHub error for %s: %s", self.repository_path, err)
            self.setup_error = True
            return

        self.name = repository.get(CONF_NAME, repo.name)
        self.available = False
        self.latest_commit_message = None
        self.latest_commit_sha = None
        self.latest_release_url = None
        self.open_issue_count = None
        self.latest_open_issue_url = None
        self.pull_request_count = None
        self.latest_open_pr_url = None
        self.stargazers = None
        self.forks = None
        self.clones = None
        self.clones_unique = None
        self.views = None
        self.views_unique = None

    def update(self):
        """Update GitHub Sensor."""
        try:
            repo = self._github_obj.get_repo(self.repository_path)

            self.stargazers = repo.stargazers_count
            self.forks = repo.forks_count

            open_pull_requests = repo.get_pulls(state="open", sort="created")
            if open_pull_requests is not None:
                self.pull_request_count = open_pull_requests.totalCount
                if open_pull_requests.totalCount > 0:
                    self.latest_open_pr_url = open_pull_requests[0].html_url

            open_issues = repo.get_issues(state="open", sort="created")
            if open_issues is not None:
                if self.pull_request_count is None:
                    self.open_issue_count = open_issues.totalCount
                else:
                    # pull requests are treated as issues too so we need to reduce the received count
                    self.open_issue_count = (
                        open_issues.totalCount - self.pull_request_count
                    )

                if open_issues.totalCount > 0:
                    self.latest_open_issue_url = open_issues[0].html_url

            latest_commit = repo.get_commits()[0]
            self.latest_commit_sha = latest_commit.sha
            self.latest_commit_message = latest_commit.commit.message

            releases = repo.get_releases()
            if releases and releases.totalCount > 0:
                self.latest_release_url = releases[0].html_url

            if repo.permissions.push:
                clones = repo.get_clones_traffic()
                if clones is not None:
                    self.clones = clones.get("count")
                    self.clones_unique = clones.get("uniques")

                views = repo.get_views_traffic()
                if views is not None:
                    self.views = views.get("count")
                    self.views_unique = views.get("uniques")

            self.available = True
        except self._github.GithubException as err:
            _LOGGER.error("GitHub error for %s: %s", self.repository_path, err)
            self.available = False



"""Support to interact with a AIS Player."""
import logging
from typing import Optional

from homeassistant.components.media_player import (
    SUPPORT_PAUSE,
    SUPPORT_PLAY,
    SUPPORT_PLAY_MEDIA,
    SUPPORT_SEEK,
    SUPPORT_SELECT_SOUND_MODE,
    SUPPORT_SHUFFLE_SET,
    SUPPORT_STOP,
    MediaPlayerEntity,
)
from homeassistant.components.media_player.const import (
    SUPPORT_BROWSE_MEDIA,
    SUPPORT_VOLUME_MUTE,
    SUPPORT_VOLUME_SET,
    SUPPORT_VOLUME_STEP,
)
from homeassistant.const import STATE_IDLE, STATE_OFF, STATE_PAUSED, STATE_PLAYING
import homeassistant.util.dt as dt_util

from .const import DOMAIN
from .media_browser import async_browse_media

_LOGGER = logging.getLogger(__name__)

SUPPORT_AIS = (
    SUPPORT_PAUSE
    | SUPPORT_PLAY_MEDIA
    | SUPPORT_PLAY
    | SUPPORT_STOP
    | SUPPORT_VOLUME_SET
    | SUPPORT_VOLUME_MUTE
    | SUPPORT_VOLUME_STEP
    | SUPPORT_SEEK
    | SUPPORT_SELECT_SOUND_MODE
    | SUPPORT_SHUFFLE_SET
    | SUPPORT_BROWSE_MEDIA
)


async def async_setup_entry(hass, config_entry, async_add_entities):
    """Set up the AIS Player."""
    _LOGGER.debug("AI-Speaker player, async_setup_entry")
    ais_gate_instance = hass.data[DOMAIN][config_entry.entry_id]
    ais_player = AisPlayerDevice(ais_gate_instance)
    async_add_entities([ais_player], True)


class AisPlayerDevice(MediaPlayerEntity):
    """Representation of a AIS Player ."""

    def turn_off(self):
        """Turn off the player."""

    def clear_playlist(self):
        """Clear the media player playlist."""

    def set_repeat(self, repeat):
        """Set the repeat for the AIS player."""

    def turn_on(self):
        """Turn on the player."""

    def __init__(self, ais_gate_instance):
        """Initialize the Ais Player device."""
        self._ais_gate = ais_gate_instance
        self._ais_id = None
        self._ais_ws_url = None
        self._ais_product = None
        self._ais_manufacturer = None
        self._ais_model = None
        self._ais_os_version = None
        self._status = None
        self._playing = False
        self._stream_image = None
        self._media_title = None
        self._media_source = None
        self._album_name = None
        self._sound_mode = "NORMAL"
        self._media_status_received_time = None
        self._media_position = 0
        self._duration = 0
        self._media_id = None
        self._volume_level = 0.5
        self._shuffle = False

    @property
    def device_info(self):
        """Device info."""
        return {
            "identifiers": {(DOMAIN, self._ais_id)},
            "name": "AI-Speaker player",
            "manufacturer": self._ais_manufacturer,
            "model": self._ais_model,
            "sw_version": self._ais_os_version,
            "via_device": None,
        }

    async def async_set_volume_level(self, volume):
        """Set the volume level."""
        self._volume_level = volume
        vol = int(volume * 100)
        await self._ais_gate.command("setVolume", vol)

    def select_sound_mode(self, sound_mode):
        """Set the sound mode."""
        self._sound_mode = sound_mode

    async def async_fetch_status(self):
        """Fetch status from AIS."""
        ais_info = await self._ais_gate.get_gate_info()
        if ais_info is None:
            _LOGGER.warning("Problem to fetch status from AI Speaker %s", ais_info)
            return
        self._ais_id = ais_info["ais_id"]
        self._ais_ws_url = ais_info["ais_url"]
        self._ais_product = ais_info.get("Product", "Player")
        self._ais_manufacturer = ais_info.get("Manufacturer", "AIS")
        self._ais_model = ais_info.get("Model", "Speaker")

        audio_status = await self._ais_gate.get_audio_status()
        if audio_status is not None:
            self._volume_level = audio_status.get("currentVolume", 100) / 100
            self._status = audio_status.get("currentStatus", None)
            self._playing = audio_status.get("playing", False)
            self._media_position = audio_status.get("currentPosition", None)
            self._duration = audio_status.get("duration", None)
            self._stream_image = audio_status.get("media_stream_image", None)
            self._media_title = audio_status.get("currentMedia", "AI-Speaker")
            self._media_source = audio_status.get("media_source", self._media_source)
            self._album_name = audio_status.get("media_album_name", "AI-Speaker")
        else:
            _LOGGER.error("Fetch status from AIS")

    @property
    def volume_level(self):
        """Return the volume level of the Player."""
        return self._volume_level

    @property
    def media_duration(self):
        """Return the duration of current playing media in seconds."""
        return float(self._duration) / 1000

    @property
    def shuffle(self):
        """Boolean if shuffle is enabled."""
        return self._shuffle

    @property
    def sound_mode(self):
        """Return the current matched sound mode."""
        return self._sound_mode

    @property
    def sound_mode_list(self):
        """Return a list of available sound modes."""
        return [
            "NORMAL",
            "BOOST",
            "TREBLE",
            "POP",
            "ROCK",
            "CLASSIC",
            "JAZZ",
            "DANCE",
            "R&P",
        ]

    @property
    def available(self):
        """Return true if AIS Player is available and connected."""
        return True

    @property
    def name(self):
        """Return the name of the device."""
        return "AI-Speaker player"

    @property
    def media_image_url(self):
        """Return the image url of current playing media."""
        if self._stream_image == "null":
            return None

        return self._stream_image

    @property
    def state(self):
        """Return the media state."""
        if self._playing is False:
            return STATE_PAUSED
        if self._status == 1:
            return STATE_IDLE
        if self._status == 2:
            return STATE_PAUSED
        if self._status == 3:
            return STATE_PLAYING
        if self._status == 4:
            return STATE_PAUSED

        return STATE_OFF

    @property
    def media_title(self):
        """Return the title of current playing media."""
        return self._media_title

    @property
    def media_album_name(self):
        """Return album of current playing media."""
        # return self._album_name
        return ""

    @property
    def app_name(self):
        """Name of the current running app."""
        app = None
        if self._media_source is not None:
            app = self._media_source
            if self._album_name is not None:
                app = str(app) + " " + str(self._album_name)
        return app

    @property
    def supported_features(self):
        """Flag media player features that are supported."""
        return SUPPORT_AIS

    @property
    def media_content_id(self):
        """Return the media content id."""
        return self._media_id

    @property
    def media_stream_image(self):
        """Return the media content id."""
        return self._stream_image

    @property
    def media_position(self):
        """Position of current playing media in seconds."""
        position = float(self._media_position) // 1000
        return int(position)

    @property
    def device_state_attributes(self):
        """Return the specific state attributes of the player."""
        return {
            "ais_id": self._ais_id,
            "ais_ws_url": self._ais_ws_url,
            "ais_exo_version": "2.2",
        }

    @property
    def unique_id(self) -> Optional[str]:
        """Return a unique ID."""
        return self._ais_id

    @property
    def media_position_updated_at(self):
        """When was the position of the current playing media valid."""
        return self._media_status_received_time

    async def async_set_shuffle(self, shuffle):
        """Enable/disable shuffle mode."""
        self._shuffle = shuffle
        await self._ais_gate.command("setPlayerShuffle", self._shuffle)

    async def async_media_seek(self, position):
        """Seek the media to a specific location."""
        if position == 0:
            val = -5000
        elif position == 1:
            val = 5000
        else:
            val = position * 1000
            self._media_status_received_time = dt_util.utcnow()
            self._media_position = position * 1000
        await self._ais_gate.command("seekTo", val)

    async def async_volume_up(self):
        """Service to send the exo the command for volume up."""
        self._volume_level = min(self._volume_level + 0.0667, 1)
        await self._ais_gate.command("upVolume", True)

    async def async_volume_down(self):
        """Service to send the exo the command for volume down."""
        self._volume_level = max(self._volume_level - 0.0667, 0)
        await self._ais_gate.command("downVolume", True)

    async def async_mute_volume(self, mute):
        """Service to send the exo the command for mute."""
        await self._ais_gate.command("setVolume", 0)

    async def async_update(self):
        """Get the latest data and update the state."""
        await self.async_fetch_status()

    async def async_media_play(self):
        """Service to send the AIS Player the command for play/pause."""
        await self._ais_gate.command("pauseAudio", False)
        self._playing = True
        self._status = 3

    async def async_media_pause(self):
        """Service to send the AIS Player the command for play/pause."""
        await self._ais_gate.command("pauseAudio", True)
        self._playing = False

    async def async_media_stop(self):
        """Service to send the AIS Player the command for stop."""
        await self._ais_gate.command("pauseAudio", True)
        self._playing = False

    async def async_media_next_track(self):
        """Service to send the AIS Player the command for next track."""
        await self._ais_gate.command("playNext", False)

    async def async_media_previous_track(self):
        """Service to send the AIS Player the command for previous track."""
        await self._ais_gate.command("playPrev", False)

    async def async_play_media(self, media_type, media_id, **kwargs):
        """Send the media player the command for playing a media."""
        self._media_id = await self._ais_gate.get_media_content_id_form_ais(media_id)
        self._media_position = 0
        self._media_status_received_time = dt_util.utcnow()
        await self._ais_gate.command("playAudio", self._media_id)

    async def async_browse_media(self, media_content_type=None, media_content_id=None):
        """Implement the media browsing helper."""
        result = await async_browse_media(
            media_content_type,
            media_content_id,
            self._ais_gate,
        )
        await self._ais_gate.cache_browse_media(result.as_dict())
        return result



"""SQLAlchemy util functions."""
from __future__ import annotations

from collections.abc import Generator
from contextlib import contextmanager
from datetime import timedelta
import logging
import os
import time

from sqlalchemy.exc import OperationalError, SQLAlchemyError
from sqlalchemy.orm.session import Session

from homeassistant.helpers.typing import HomeAssistantType
import homeassistant.util.dt as dt_util

from .const import CONF_DB_INTEGRITY_CHECK, DATA_INSTANCE, SQLITE_URL_PREFIX
from .models import ALL_TABLES, process_timestamp

_LOGGER = logging.getLogger(__name__)

RETRIES = 3
QUERY_RETRY_WAIT = 0.1
SQLITE3_POSTFIXES = ["", "-wal", "-shm"]

# This is the maximum time after the recorder ends the session
# before we no longer consider startup to be a "restart" and we
# should do a check on the sqlite3 database.
MAX_RESTART_TIME = timedelta(minutes=10)


@contextmanager
def session_scope(
    *, hass: HomeAssistantType | None = None, session: Session | None = None
) -> Generator[Session, None, None]:
    """Provide a transactional scope around a series of operations."""
    if session is None and hass is not None:
        session = hass.data[DATA_INSTANCE].get_session()

    if session is None:
        raise RuntimeError("Session required")

    need_rollback = False
    try:
        yield session
        if session.transaction:
            need_rollback = True
            session.commit()
    except Exception as err:
        _LOGGER.error("Error executing query: %s", err)
        if need_rollback:
            session.rollback()
        raise
    finally:
        session.close()


def commit(session, work):
    """Commit & retry work: Either a model or in a function."""
    for _ in range(0, RETRIES):
        try:
            if callable(work):
                work(session)
            else:
                session.add(work)
            session.commit()
            return True
        except OperationalError as err:
            _LOGGER.error("Error executing query: %s", err)
            session.rollback()
            time.sleep(QUERY_RETRY_WAIT)
    return False


def execute(qry, to_native=False, validate_entity_ids=True):
    """Query the database and convert the objects to HA native form.

    This method also retries a few times in the case of stale connections.
    """

    for tryno in range(0, RETRIES):
        try:
            timer_start = time.perf_counter()
            if to_native:
                result = [
                    row
                    for row in (
                        row.to_native(validate_entity_id=validate_entity_ids)
                        for row in qry
                    )
                    if row is not None
                ]
            else:
                result = list(qry)

            if _LOGGER.isEnabledFor(logging.DEBUG):
                elapsed = time.perf_counter() - timer_start
                if to_native:
                    _LOGGER.debug(
                        "converting %d rows to native objects took %fs",
                        len(result),
                        elapsed,
                    )
                else:
                    _LOGGER.debug(
                        "querying %d rows took %fs",
                        len(result),
                        elapsed,
                    )

            return result
        except SQLAlchemyError as err:
            _LOGGER.error("Error executing query: %s", err)

            if tryno == RETRIES - 1:
                raise
            time.sleep(QUERY_RETRY_WAIT)


def validate_or_move_away_sqlite_database(dburl: str, db_integrity_check: bool) -> bool:
    """Ensure that the database is valid or move it away."""
    dbpath = dburl_to_path(dburl)

    if not os.path.exists(dbpath):
        # Database does not exist yet, this is OK
        return True

    if not validate_sqlite_database(dbpath, db_integrity_check):
        move_away_broken_database(dbpath)
        return False

    return True


def dburl_to_path(dburl):
    """Convert the db url into a filesystem path."""
    return dburl[len(SQLITE_URL_PREFIX) :]


def last_run_was_recently_clean(cursor):
    """Verify the last recorder run was recently clean."""

    cursor.execute("SELECT end FROM recorder_runs ORDER BY start DESC LIMIT 1;")
    end_time = cursor.fetchone()

    if not end_time or not end_time[0]:
        return False

    last_run_end_time = process_timestamp(dt_util.parse_datetime(end_time[0]))
    now = dt_util.utcnow()

    _LOGGER.debug("The last run ended at: %s (now: %s)", last_run_end_time, now)

    if last_run_end_time + MAX_RESTART_TIME < now:
        return False

    return True


def basic_sanity_check(cursor):
    """Check tables to make sure select does not fail."""

    for table in ALL_TABLES:
        cursor.execute(f"SELECT * FROM {table} LIMIT 1;")  # nosec # not injection

    return True


def validate_sqlite_database(dbpath: str, db_integrity_check: bool) -> bool:
    """Run a quick check on an sqlite database to see if it is corrupt."""
    import sqlite3  # pylint: disable=import-outside-toplevel

    try:
        conn = sqlite3.connect(dbpath)
        run_checks_on_open_db(dbpath, conn.cursor(), db_integrity_check)
        conn.close()
    except sqlite3.DatabaseError:
        _LOGGER.exception("The database at %s is corrupt or malformed", dbpath)
        return False

    return True


def run_checks_on_open_db(dbpath, cursor, db_integrity_check):
    """Run checks that will generate a sqlite3 exception if there is corruption."""
    sanity_check_passed = basic_sanity_check(cursor)
    last_run_was_clean = last_run_was_recently_clean(cursor)

    if sanity_check_passed and last_run_was_clean:
        _LOGGER.debug(
            "The quick_check will be skipped as the system was restarted cleanly and passed the basic sanity check"
        )
        return

    if not db_integrity_check:
        # Always warn so when it does fail they remember it has
        # been manually disabled
        _LOGGER.warning(
            "The quick_check on the sqlite3 database at %s was skipped because %s was disabled",
            dbpath,
            CONF_DB_INTEGRITY_CHECK,
        )
        return

    if not sanity_check_passed:
        _LOGGER.warning(
            "The database sanity check failed to validate the sqlite3 database at %s",
            dbpath,
        )

    if not last_run_was_clean:
        _LOGGER.warning(
            "The system could not validate that the sqlite3 database at %s was shutdown cleanly",
            dbpath,
        )

    _LOGGER.info(
        "A quick_check is being performed on the sqlite3 database at %s", dbpath
    )
    cursor.execute("PRAGMA QUICK_CHECK")


def move_away_broken_database(dbfile: str) -> None:
    """Move away a broken sqlite3 database."""

    isotime = dt_util.utcnow().isoformat()
    corrupt_postfix = f".corrupt.{isotime}"

    _LOGGER.error(
        "The system will rename the corrupt database file %s to %s in order to allow startup to proceed",
        dbfile,
        f"{dbfile}{corrupt_postfix}",
    )

    for postfix in SQLITE3_POSTFIXES:
        path = f"{dbfile}{postfix}"
        if not os.path.exists(path):
            continue
        os.rename(path, f"{path}{corrupt_postfix}")



"""The zodiac component."""
import voluptuous as vol

from homeassistant.core import HomeAssistant
from homeassistant.helpers.discovery import async_load_platform

from .const import DOMAIN

CONFIG_SCHEMA = vol.Schema(
    {vol.Optional(DOMAIN): {}},
    extra=vol.ALLOW_EXTRA,
)


async def async_setup(hass: HomeAssistant, config: dict):
    """Set up the zodiac component."""
    hass.async_create_task(async_load_platform(hass, "sensor", DOMAIN, {}, config))

    return True



"""Support for a ScreenLogic Binary Sensor."""
import logging

from screenlogicpy.const import DEVICE_TYPE, ON_OFF

from homeassistant.components.binary_sensor import (
    DEVICE_CLASS_PROBLEM,
    BinarySensorEntity,
)

from . import ScreenlogicEntity
from .const import DOMAIN

_LOGGER = logging.getLogger(__name__)

SL_DEVICE_TYPE_TO_HA_DEVICE_CLASS = {DEVICE_TYPE.ALARM: DEVICE_CLASS_PROBLEM}


async def async_setup_entry(hass, config_entry, async_add_entities):
    """Set up entry."""
    entities = []
    data = hass.data[DOMAIN][config_entry.entry_id]
    coordinator = data["coordinator"]

    for binary_sensor in data["devices"]["binary_sensor"]:
        entities.append(ScreenLogicBinarySensor(coordinator, binary_sensor))
    async_add_entities(entities)


class ScreenLogicBinarySensor(ScreenlogicEntity, BinarySensorEntity):
    """Representation of a ScreenLogic binary sensor entity."""

    @property
    def name(self):
        """Return the sensor name."""
        return f"{self.gateway_name} {self.sensor['name']}"

    @property
    def device_class(self):
        """Return the device class."""
        device_class = self.sensor.get("device_type")
        return SL_DEVICE_TYPE_TO_HA_DEVICE_CLASS.get(device_class)

    @property
    def is_on(self) -> bool:
        """Determine if the sensor is on."""
        return self.sensor["value"] == ON_OFF.ON

    @property
    def sensor(self):
        """Shortcut to access the sensor data."""
        return self.coordinator.data["sensors"][self._data_key]



"""Support for Sensibo wifi-enabled home thermostats."""

import asyncio
import logging

import aiohttp
import async_timeout
import pysensibo
import voluptuous as vol

from homeassistant.components.climate import PLATFORM_SCHEMA, ClimateEntity
from homeassistant.components.climate.const import (
    HVAC_MODE_COOL,
    HVAC_MODE_DRY,
    HVAC_MODE_FAN_ONLY,
    HVAC_MODE_HEAT,
    HVAC_MODE_HEAT_COOL,
    HVAC_MODE_OFF,
    SUPPORT_FAN_MODE,
    SUPPORT_SWING_MODE,
    SUPPORT_TARGET_TEMPERATURE,
)
from homeassistant.const import (
    ATTR_ENTITY_ID,
    ATTR_STATE,
    ATTR_TEMPERATURE,
    CONF_API_KEY,
    CONF_ID,
    STATE_ON,
    TEMP_CELSIUS,
    TEMP_FAHRENHEIT,
)
from homeassistant.exceptions import PlatformNotReady
from homeassistant.helpers import config_validation as cv
from homeassistant.helpers.aiohttp_client import async_get_clientsession
from homeassistant.util.temperature import convert as convert_temperature

from .const import DOMAIN as SENSIBO_DOMAIN

_LOGGER = logging.getLogger(__name__)

ALL = ["all"]
TIMEOUT = 10

SERVICE_ASSUME_STATE = "assume_state"

PLATFORM_SCHEMA = PLATFORM_SCHEMA.extend(
    {
        vol.Required(CONF_API_KEY): cv.string,
        vol.Optional(CONF_ID, default=ALL): vol.All(cv.ensure_list, [cv.string]),
    }
)

ASSUME_STATE_SCHEMA = vol.Schema(
    {vol.Optional(ATTR_ENTITY_ID): cv.entity_ids, vol.Required(ATTR_STATE): cv.string}
)

_FETCH_FIELDS = ",".join(
    [
        "room{name}",
        "measurements",
        "remoteCapabilities",
        "acState",
        "connectionStatus{isAlive}",
        "temperatureUnit",
    ]
)
_INITIAL_FETCH_FIELDS = f"id,{_FETCH_FIELDS}"

FIELD_TO_FLAG = {
    "fanLevel": SUPPORT_FAN_MODE,
    "swing": SUPPORT_SWING_MODE,
    "targetTemperature": SUPPORT_TARGET_TEMPERATURE,
}

SENSIBO_TO_HA = {
    "cool": HVAC_MODE_COOL,
    "heat": HVAC_MODE_HEAT,
    "fan": HVAC_MODE_FAN_ONLY,
    "auto": HVAC_MODE_HEAT_COOL,
    "dry": HVAC_MODE_DRY,
}

HA_TO_SENSIBO = {value: key for key, value in SENSIBO_TO_HA.items()}


async def async_setup_platform(hass, config, async_add_entities, discovery_info=None):
    """Set up Sensibo devices."""
    client = pysensibo.SensiboClient(
        config[CONF_API_KEY], session=async_get_clientsession(hass), timeout=TIMEOUT
    )
    devices = []
    try:
        for dev in await client.async_get_devices(_INITIAL_FETCH_FIELDS):
            if config[CONF_ID] == ALL or dev["id"] in config[CONF_ID]:
                devices.append(
                    SensiboClimate(client, dev, hass.config.units.temperature_unit)
                )
    except (
        aiohttp.client_exceptions.ClientConnectorError,
        asyncio.TimeoutError,
        pysensibo.SensiboError,
    ) as err:
        _LOGGER.exception("Failed to connect to Sensibo servers")
        raise PlatformNotReady from err

    if not devices:
        return

    async_add_entities(devices)

    async def async_assume_state(service):
        """Set state according to external service call.."""
        entity_ids = service.data.get(ATTR_ENTITY_ID)
        if entity_ids:
            target_climate = [
                device for device in devices if device.entity_id in entity_ids
            ]
        else:
            target_climate = devices

        update_tasks = []
        for climate in target_climate:
            await climate.async_assume_state(service.data.get(ATTR_STATE))
            update_tasks.append(climate.async_update_ha_state(True))

        if update_tasks:
            await asyncio.wait(update_tasks)

    hass.services.async_register(
        SENSIBO_DOMAIN,
        SERVICE_ASSUME_STATE,
        async_assume_state,
        schema=ASSUME_STATE_SCHEMA,
    )


class SensiboClimate(ClimateEntity):
    """Representation of a Sensibo device."""

    def __init__(self, client, data, units):
        """Build SensiboClimate.

        client: aiohttp session.
        data: initially-fetched data.
        """
        self._client = client
        self._id = data["id"]
        self._external_state = None
        self._units = units
        self._available = False
        self._do_update(data)

    @property
    def supported_features(self):
        """Return the list of supported features."""
        return self._supported_features

    def _do_update(self, data):
        self._name = data["room"]["name"]
        self._measurements = data["measurements"]
        self._ac_states = data["acState"]
        self._available = data["connectionStatus"]["isAlive"]
        capabilities = data["remoteCapabilities"]
        self._operations = [SENSIBO_TO_HA[mode] for mode in capabilities["modes"]]
        self._operations.append(HVAC_MODE_OFF)
        self._current_capabilities = capabilities["modes"][self._ac_states["mode"]]
        temperature_unit_key = data.get("temperatureUnit") or self._ac_states.get(
            "temperatureUnit"
        )
        if temperature_unit_key:
            self._temperature_unit = (
                TEMP_CELSIUS if temperature_unit_key == "C" else TEMP_FAHRENHEIT
            )
            self._temperatures_list = (
                self._current_capabilities["temperatures"]
                .get(temperature_unit_key, {})
                .get("values", [])
            )
        else:
            self._temperature_unit = self._units
            self._temperatures_list = []
        self._supported_features = 0
        for key in self._ac_states:
            if key in FIELD_TO_FLAG:
                self._supported_features |= FIELD_TO_FLAG[key]

    @property
    def state(self):
        """Return the current state."""
        return self._external_state or super().state

    @property
    def extra_state_attributes(self):
        """Return the state attributes."""
        return {"battery": self.current_battery}

    @property
    def temperature_unit(self):
        """Return the unit of measurement which this thermostat uses."""
        return self._temperature_unit

    @property
    def available(self):
        """Return True if entity is available."""
        return self._available

    @property
    def target_temperature(self):
        """Return the temperature we try to reach."""
        return self._ac_states.get("targetTemperature")

    @property
    def target_temperature_step(self):
        """Return the supported step of target temperature."""
        if self.temperature_unit == self.hass.config.units.temperature_unit:
            # We are working in same units as the a/c unit. Use whole degrees
            # like the API supports.
            return 1
        # Unit conversion is going on. No point to stick to specific steps.
        return None

    @property
    def hvac_mode(self):
        """Return current operation ie. heat, cool, idle."""
        if not self._ac_states["on"]:
            return HVAC_MODE_OFF
        return SENSIBO_TO_HA.get(self._ac_states["mode"])

    @property
    def current_humidity(self):
        """Return the current humidity."""
        return self._measurements["humidity"]

    @property
    def current_battery(self):
        """Return the current battery voltage."""
        return self._measurements.get("batteryVoltage")

    @property
    def current_temperature(self):
        """Return the current temperature."""
        # This field is not affected by temperatureUnit.
        # It is always in C
        return convert_temperature(
            self._measurements["temperature"], TEMP_CELSIUS, self.temperature_unit
        )

    @property
    def hvac_modes(self):
        """List of available operation modes."""
        return self._operations

    @property
    def fan_mode(self):
        """Return the fan setting."""
        return self._ac_states.get("fanLevel")

    @property
    def fan_modes(self):
        """List of available fan modes."""
        return self._current_capabilities.get("fanLevels")

    @property
    def swing_mode(self):
        """Return the fan setting."""
        return self._ac_states.get("swing")

    @property
    def swing_modes(self):
        """List of available swing modes."""
        return self._current_capabilities.get("swing")

    @property
    def name(self):
        """Return the name of the entity."""
        return self._name

    @property
    def min_temp(self):
        """Return the minimum temperature."""
        return (
            self._temperatures_list[0] if self._temperatures_list else super().min_temp
        )

    @property
    def max_temp(self):
        """Return the maximum temperature."""
        return (
            self._temperatures_list[-1] if self._temperatures_list else super().max_temp
        )

    @property
    def unique_id(self):
        """Return unique ID based on Sensibo ID."""
        return self._id

    async def async_set_temperature(self, **kwargs):
        """Set new target temperature."""
        temperature = kwargs.get(ATTR_TEMPERATURE)
        if temperature is None:
            return
        temperature = int(temperature)
        if temperature not in self._temperatures_list:
            # Requested temperature is not supported.
            if temperature == self.target_temperature:
                return
            index = self._temperatures_list.index(self.target_temperature)
            if (
                temperature > self.target_temperature
                and index < len(self._temperatures_list) - 1
            ):
                temperature = self._temperatures_list[index + 1]
            elif temperature < self.target_temperature and index > 0:
                temperature = self._temperatures_list[index - 1]
            else:
                return

        with async_timeout.timeout(TIMEOUT):
            await self._client.async_set_ac_state_property(
                self._id, "targetTemperature", temperature, self._ac_states
            )

    async def async_set_fan_mode(self, fan_mode):
        """Set new target fan mode."""
        with async_timeout.timeout(TIMEOUT):
            await self._client.async_set_ac_state_property(
                self._id, "fanLevel", fan_mode, self._ac_states
            )

    async def async_set_hvac_mode(self, hvac_mode):
        """Set new target operation mode."""
        if hvac_mode == HVAC_MODE_OFF:
            with async_timeout.timeout(TIMEOUT):
                await self._client.async_set_ac_state_property(
                    self._id, "on", False, self._ac_states
                )
            return

        # Turn on if not currently on.
        if not self._ac_states["on"]:
            with async_timeout.timeout(TIMEOUT):
                await self._client.async_set_ac_state_property(
                    self._id, "on", True, self._ac_states
                )

        with async_timeout.timeout(TIMEOUT):
            await self._client.async_set_ac_state_property(
                self._id, "mode", HA_TO_SENSIBO[hvac_mode], self._ac_states
            )

    async def async_set_swing_mode(self, swing_mode):
        """Set new target swing operation."""
        with async_timeout.timeout(TIMEOUT):
            await self._client.async_set_ac_state_property(
                self._id, "swing", swing_mode, self._ac_states
            )

    async def async_turn_on(self):
        """Turn Sensibo unit on."""
        with async_timeout.timeout(TIMEOUT):
            await self._client.async_set_ac_state_property(
                self._id, "on", True, self._ac_states
            )

    async def async_turn_off(self):
        """Turn Sensibo unit on."""
        with async_timeout.timeout(TIMEOUT):
            await self._client.async_set_ac_state_property(
                self._id, "on", False, self._ac_states
            )

    async def async_assume_state(self, state):
        """Set external state."""
        change_needed = (state != HVAC_MODE_OFF and not self._ac_states["on"]) or (
            state == HVAC_MODE_OFF and self._ac_states["on"]
        )

        if change_needed:
            with async_timeout.timeout(TIMEOUT):
                await self._client.async_set_ac_state_property(
                    self._id,
                    "on",
                    state != HVAC_MODE_OFF,  # value
                    self._ac_states,
                    True,  # assumed_state
                )

        if state in [STATE_ON, HVAC_MODE_OFF]:
            self._external_state = None
        else:
            self._external_state = state

    async def async_update(self):
        """Retrieve latest state."""
        try:
            with async_timeout.timeout(TIMEOUT):
                data = await self._client.async_get_device(self._id, _FETCH_FIELDS)
                self._do_update(data)
        except (aiohttp.client_exceptions.ClientError, pysensibo.SensiboError):
            _LOGGER.warning("Failed to connect to Sensibo servers")
            self._available = False



"""Alexa state report code."""
from __future__ import annotations

import asyncio
import json
import logging

import aiohttp
import async_timeout

from homeassistant.const import HTTP_ACCEPTED, MATCH_ALL, STATE_ON
from homeassistant.core import HomeAssistant, State, callback
from homeassistant.helpers.significant_change import create_checker
import homeassistant.util.dt as dt_util

from .const import API_CHANGE, DOMAIN, Cause
from .entities import ENTITY_ADAPTERS, AlexaEntity, generate_alexa_id
from .messages import AlexaResponse

_LOGGER = logging.getLogger(__name__)
DEFAULT_TIMEOUT = 10


async def async_enable_proactive_mode(hass, smart_home_config):
    """Enable the proactive mode.

    Proactive mode makes this component report state changes to Alexa.
    """
    # Validate we can get access token.
    await smart_home_config.async_get_access_token()

    @callback
    def extra_significant_check(
        hass: HomeAssistant,
        old_state: str,
        old_attrs: dict,
        old_extra_arg: dict,
        new_state: str,
        new_attrs: dict,
        new_extra_arg: dict,
    ):
        """Check if the serialized data has changed."""
        return old_extra_arg is not None and old_extra_arg != new_extra_arg

    checker = await create_checker(hass, DOMAIN, extra_significant_check)

    async def async_entity_state_listener(
        changed_entity: str,
        old_state: State | None,
        new_state: State | None,
    ):
        if not hass.is_running:
            return

        if not new_state:
            return

        if new_state.domain not in ENTITY_ADAPTERS:
            return

        if not smart_home_config.should_expose(changed_entity):
            _LOGGER.debug("Not exposing %s because filtered by config", changed_entity)
            return

        alexa_changed_entity: AlexaEntity = ENTITY_ADAPTERS[new_state.domain](
            hass, smart_home_config, new_state
        )

        # Determine how entity should be reported on
        should_report = False
        should_doorbell = False

        for interface in alexa_changed_entity.interfaces():
            if not should_report and interface.properties_proactively_reported():
                should_report = True

            if interface.name() == "Alexa.DoorbellEventSource":
                should_doorbell = True
                break

        if not should_report and not should_doorbell:
            return

        if should_doorbell:
            if new_state.state == STATE_ON:
                await async_send_doorbell_event_message(
                    hass, smart_home_config, alexa_changed_entity
                )
            return

        alexa_properties = list(alexa_changed_entity.serialize_properties())

        if not checker.async_is_significant_change(
            new_state, extra_arg=alexa_properties
        ):
            return

        await async_send_changereport_message(
            hass, smart_home_config, alexa_changed_entity, alexa_properties
        )

    return hass.helpers.event.async_track_state_change(
        MATCH_ALL, async_entity_state_listener
    )


async def async_send_changereport_message(
    hass, config, alexa_entity, alexa_properties, *, invalidate_access_token=True
):
    """Send a ChangeReport message for an Alexa entity.

    https://developer.amazon.com/docs/smarthome/state-reporting-for-a-smart-home-skill.html#report-state-with-changereport-events
    """
    token = await config.async_get_access_token()

    headers = {"Authorization": f"Bearer {token}"}

    endpoint = alexa_entity.alexa_id()

    payload = {
        API_CHANGE: {
            "cause": {"type": Cause.APP_INTERACTION},
            "properties": alexa_properties,
        }
    }

    message = AlexaResponse(name="ChangeReport", namespace="Alexa", payload=payload)
    message.set_endpoint_full(token, endpoint)

    message_serialized = message.serialize()
    session = hass.helpers.aiohttp_client.async_get_clientsession()

    try:
        with async_timeout.timeout(DEFAULT_TIMEOUT):
            response = await session.post(
                config.endpoint,
                headers=headers,
                json=message_serialized,
                allow_redirects=True,
            )

    except (asyncio.TimeoutError, aiohttp.ClientError):
        _LOGGER.error("Timeout sending report to Alexa")
        return

    response_text = await response.text()

    _LOGGER.debug("Sent: %s", json.dumps(message_serialized))
    _LOGGER.debug("Received (%s): %s", response.status, response_text)

    if response.status == HTTP_ACCEPTED:
        return

    response_json = json.loads(response_text)

    if (
        response_json["payload"]["code"] == "INVALID_ACCESS_TOKEN_EXCEPTION"
        and not invalidate_access_token
    ):
        config.async_invalidate_access_token()
        return await async_send_changereport_message(
            hass, config, alexa_entity, alexa_properties, invalidate_access_token=False
        )

    _LOGGER.error(
        "Error when sending ChangeReport to Alexa: %s: %s",
        response_json["payload"]["code"],
        response_json["payload"]["description"],
    )


async def async_send_add_or_update_message(hass, config, entity_ids):
    """Send an AddOrUpdateReport message for entities.

    https://developer.amazon.com/docs/device-apis/alexa-discovery.html#add-or-update-report
    """
    token = await config.async_get_access_token()

    headers = {"Authorization": f"Bearer {token}"}

    endpoints = []

    for entity_id in entity_ids:
        domain = entity_id.split(".", 1)[0]

        if domain not in ENTITY_ADAPTERS:
            continue

        alexa_entity = ENTITY_ADAPTERS[domain](hass, config, hass.states.get(entity_id))
        endpoints.append(alexa_entity.serialize_discovery())

    payload = {"endpoints": endpoints, "scope": {"type": "BearerToken", "token": token}}

    message = AlexaResponse(
        name="AddOrUpdateReport", namespace="Alexa.Discovery", payload=payload
    )

    message_serialized = message.serialize()
    session = hass.helpers.aiohttp_client.async_get_clientsession()

    return await session.post(
        config.endpoint, headers=headers, json=message_serialized, allow_redirects=True
    )


async def async_send_delete_message(hass, config, entity_ids):
    """Send an DeleteReport message for entities.

    https://developer.amazon.com/docs/device-apis/alexa-discovery.html#deletereport-event
    """
    token = await config.async_get_access_token()

    headers = {"Authorization": f"Bearer {token}"}

    endpoints = []

    for entity_id in entity_ids:
        domain = entity_id.split(".", 1)[0]

        if domain not in ENTITY_ADAPTERS:
            continue

        endpoints.append({"endpointId": generate_alexa_id(entity_id)})

    payload = {"endpoints": endpoints, "scope": {"type": "BearerToken", "token": token}}

    message = AlexaResponse(
        name="DeleteReport", namespace="Alexa.Discovery", payload=payload
    )

    message_serialized = message.serialize()
    session = hass.helpers.aiohttp_client.async_get_clientsession()

    return await session.post(
        config.endpoint, headers=headers, json=message_serialized, allow_redirects=True
    )


async def async_send_doorbell_event_message(hass, config, alexa_entity):
    """Send a DoorbellPress event message for an Alexa entity.

    https://developer.amazon.com/en-US/docs/alexa/device-apis/alexa-doorbelleventsource.html
    """
    token = await config.async_get_access_token()

    headers = {"Authorization": f"Bearer {token}"}

    endpoint = alexa_entity.alexa_id()

    message = AlexaResponse(
        name="DoorbellPress",
        namespace="Alexa.DoorbellEventSource",
        payload={
            "cause": {"type": Cause.PHYSICAL_INTERACTION},
            "timestamp": f"{dt_util.utcnow().replace(tzinfo=None).isoformat()}Z",
        },
    )

    message.set_endpoint_full(token, endpoint)

    message_serialized = message.serialize()
    session = hass.helpers.aiohttp_client.async_get_clientsession()

    try:
        with async_timeout.timeout(DEFAULT_TIMEOUT):
            response = await session.post(
                config.endpoint,
                headers=headers,
                json=message_serialized,
                allow_redirects=True,
            )

    except (asyncio.TimeoutError, aiohttp.ClientError):
        _LOGGER.error("Timeout sending report to Alexa")
        return

    response_text = await response.text()

    _LOGGER.debug("Sent: %s", json.dumps(message_serialized))
    _LOGGER.debug("Received (%s): %s", response.status, response_text)

    if response.status == HTTP_ACCEPTED:
        return

    response_json = json.loads(response_text)

    _LOGGER.error(
        "Error when sending DoorbellPress event to Alexa: %s: %s",
        response_json["payload"]["code"],
        response_json["payload"]["description"],
    )



"""Support for Rituals Perfume Genie switches."""
from datetime import timedelta
import logging

import aiohttp

from homeassistant.components.switch import SwitchEntity

from .const import DOMAIN

_LOGGER = logging.getLogger(__name__)

SCAN_INTERVAL = timedelta(seconds=30)

ON_STATE = "1"
AVAILABLE_STATE = 1

MANUFACTURER = "Rituals Cosmetics"
MODEL = "Diffuser"
ICON = "mdi:fan"


async def async_setup_entry(hass, config_entry, async_add_entities):
    """Set up the diffuser switch."""
    account = hass.data[DOMAIN][config_entry.entry_id]
    diffusers = await account.get_devices()

    entities = []
    for diffuser in diffusers:
        entities.append(DiffuserSwitch(diffuser))

    async_add_entities(entities, True)


class DiffuserSwitch(SwitchEntity):
    """Representation of a diffuser switch."""

    def __init__(self, diffuser):
        """Initialize the switch."""
        self._diffuser = diffuser
        self._available = True

    @property
    def device_info(self):
        """Return information about the device."""
        return {
            "name": self._diffuser.data["hub"]["attributes"]["roomnamec"],
            "identifiers": {(DOMAIN, self._diffuser.data["hub"]["hublot"])},
            "manufacturer": MANUFACTURER,
            "model": MODEL,
            "sw_version": self._diffuser.data["hub"]["sensors"]["versionc"],
        }

    @property
    def unique_id(self):
        """Return the unique ID of the device."""
        return self._diffuser.data["hub"]["hublot"]

    @property
    def available(self):
        """Return if the device is available."""
        return self._available

    @property
    def name(self):
        """Return the name of the device."""
        return self._diffuser.data["hub"]["attributes"]["roomnamec"]

    @property
    def icon(self):
        """Return the icon of the device."""
        return ICON

    @property
    def extra_state_attributes(self):
        """Return the device state attributes."""
        attributes = {
            "fan_speed": self._diffuser.data["hub"]["attributes"]["speedc"],
            "room_size": self._diffuser.data["hub"]["attributes"]["roomc"],
        }
        return attributes

    @property
    def is_on(self):
        """If the device is currently on or off."""
        return self._diffuser.data["hub"]["attributes"]["fanc"] == ON_STATE

    async def async_turn_on(self, **kwargs):
        """Turn the device on."""
        await self._diffuser.turn_on()

    async def async_turn_off(self, **kwargs):
        """Turn the device off."""
        await self._diffuser.turn_off()

    async def async_update(self):
        """Update the data of the device."""
        try:
            await self._diffuser.update_data()
        except aiohttp.ClientError:
            self._available = False
            _LOGGER.error("Unable to retrieve data from rituals.sense-company.com")
        else:
            self._available = self._diffuser.data["hub"]["status"] == AVAILABLE_STATE



"""
Support for AIS knowledge

For more details about this component, please refer to the documentation at
https://www.ai-speaker.com
"""
import logging

import voluptuous as vol

from homeassistant.components import ais_cloud
from homeassistant.components.ais_dom import ais_global
from homeassistant.helpers import config_validation as cv

aisCloud = None

DOMAIN = "ais_knowledge_service"
SERVICE_ASK = "ask"
SERVICE_ASK_WIKI = "ask_wiki"
ATTR_TEXT = "text"
ATTR_SAY_IT = "say_it"
SERVICE_ASK_SCHEMA = vol.Schema(
    {vol.Required(ATTR_TEXT): cv.string, vol.Optional(ATTR_SAY_IT): cv.boolean}
)
_LOGGER = logging.getLogger(__name__)
GKS_URL = "https://kgsearch.googleapis.com/v1/entities:search"
G_GKS_KEY = None


async def async_setup(hass, config):
    """Register the service."""
    global aisCloud
    aisCloud = ais_cloud.AisCloudWS(hass)
    await async_get_key(hass)

    async def async_ask(service):
        """ask service about info"""
        query = service.data[ATTR_TEXT]
        await async_process_ask(hass, query)

    async def async_ask_wiki(service):
        """ask wikipedia service about info"""
        query = service.data[ATTR_TEXT]
        await async_process_ask_wiki(hass, query)

    # register services
    hass.services.async_register(
        DOMAIN, SERVICE_ASK, async_ask, schema=SERVICE_ASK_SCHEMA
    )

    hass.services.async_register(
        DOMAIN, SERVICE_ASK_WIKI, async_ask_wiki, schema=SERVICE_ASK_SCHEMA
    )

    return True


async def async_process_ask(hass, query):
    import requests

    global G_GKS_KEY
    """Ask the service about text."""

    full_message = ""
    image_url = None
    if G_GKS_KEY is None:
        try:
            json_ws_resp = await aisCloud.async_key("kgsearch")
            G_GKS_KEY = json_ws_resp["key"]
        except Exception as e:
            _LOGGER.error("search error: " + str(e))
            return

    req = requests.get(
        GKS_URL,
        params={
            "query": query,
            "limit": 1,
            "indent": True,
            "key": G_GKS_KEY,
            "languages": "pl",
        },
    )
    try:
        response = req.json()
        element = response["itemListElement"][0]
        result = element["result"]
        full_message = result["description"]
        # try to get more info
        try:
            full_message = (
                full_message + "\n\n" + result["detailedDescription"]["articleBody"]
            )
        except Exception:
            pass
        # try to get image
        try:
            image_url = result["image"]["contentUrl"]
        except Exception:
            pass

    except Exception:
        full_message = "Brak wyników"

    await hass.services.async_call(
        "ais_ai_service", "say_it", {"text": full_message, "img": image_url}
    )

    return full_message


async def async_process_ask_wiki(hass, query):
    import wikipedia

    wikipedia.set_lang("pl")
    """Ask the service about text."""
    image_url = None
    try:
        full_message = wikipedia.summary(query)
    except Exception:
        full_message = "Brak wyników"

    try:
        images = wikipedia.page(query).images
        for image in images:
            if not image.endswith(".svg"):
                image_url = image

    except Exception:
        full_message = "Brak wyników"

    await hass.services.async_call(
        "ais_ai_service", "say_it", {"text": full_message, "img": image_url}
    )

    return full_message


async def async_get_key(hass):
    global G_GKS_KEY
    try:
        json_ws_resp = await aisCloud.async_key("kgsearch")
        G_GKS_KEY = json_ws_resp["key"]
    except:
        ais_global.G_OFFLINE_MODE = True



"""Jabber (XMPP) notification service."""
from concurrent.futures import TimeoutError as FutTimeoutError
import logging
import mimetypes
import pathlib
import random
import string

import requests
import slixmpp
from slixmpp.exceptions import IqError, IqTimeout, XMPPError
from slixmpp.plugins.xep_0363.http_upload import (
    FileTooBig,
    FileUploadError,
    UploadServiceNotFound,
)
from slixmpp.xmlstream.xmlstream import NotConnectedError
import voluptuous as vol

from homeassistant.components.notify import (
    ATTR_TITLE,
    ATTR_TITLE_DEFAULT,
    PLATFORM_SCHEMA,
    BaseNotificationService,
)
from homeassistant.const import (
    CONF_PASSWORD,
    CONF_RECIPIENT,
    CONF_RESOURCE,
    CONF_ROOM,
    CONF_SENDER,
    HTTP_BAD_REQUEST,
)
import homeassistant.helpers.config_validation as cv
import homeassistant.helpers.template as template_helper

_LOGGER = logging.getLogger(__name__)

ATTR_DATA = "data"
ATTR_PATH = "path"
ATTR_PATH_TEMPLATE = "path_template"
ATTR_TIMEOUT = "timeout"
ATTR_URL = "url"
ATTR_URL_TEMPLATE = "url_template"
ATTR_VERIFY = "verify"

CONF_TLS = "tls"
CONF_VERIFY = "verify"

DEFAULT_CONTENT_TYPE = "application/octet-stream"
DEFAULT_RESOURCE = "home-assistant"
XEP_0363_TIMEOUT = 10

PLATFORM_SCHEMA = PLATFORM_SCHEMA.extend(
    {
        vol.Required(CONF_SENDER): cv.string,
        vol.Required(CONF_PASSWORD): cv.string,
        vol.Required(CONF_RECIPIENT): vol.All(cv.ensure_list, [cv.string]),
        vol.Optional(CONF_RESOURCE, default=DEFAULT_RESOURCE): cv.string,
        vol.Optional(CONF_ROOM, default=""): cv.string,
        vol.Optional(CONF_TLS, default=True): cv.boolean,
        vol.Optional(CONF_VERIFY, default=True): cv.boolean,
    }
)


async def async_get_service(hass, config, discovery_info=None):
    """Get the Jabber (XMPP) notification service."""
    return XmppNotificationService(
        config.get(CONF_SENDER),
        config.get(CONF_RESOURCE),
        config.get(CONF_PASSWORD),
        config.get(CONF_RECIPIENT),
        config.get(CONF_TLS),
        config.get(CONF_VERIFY),
        config.get(CONF_ROOM),
        hass,
    )


class XmppNotificationService(BaseNotificationService):
    """Implement the notification service for Jabber (XMPP)."""

    def __init__(self, sender, resource, password, recipient, tls, verify, room, hass):
        """Initialize the service."""
        self._hass = hass
        self._sender = sender
        self._resource = resource
        self._password = password
        self._recipients = recipient
        self._tls = tls
        self._verify = verify
        self._room = room

    async def async_send_message(self, message="", **kwargs):
        """Send a message to a user."""
        title = kwargs.get(ATTR_TITLE, ATTR_TITLE_DEFAULT)
        text = f"{title}: {message}" if title else message
        data = kwargs.get(ATTR_DATA)
        timeout = data.get(ATTR_TIMEOUT, XEP_0363_TIMEOUT) if data else None

        await async_send_message(
            f"{self._sender}/{self._resource}",
            self._password,
            self._recipients,
            self._tls,
            self._verify,
            self._room,
            self._hass,
            text,
            timeout,
            data,
        )


async def async_send_message(
    sender,
    password,
    recipients,
    use_tls,
    verify_certificate,
    room,
    hass,
    message,
    timeout=None,
    data=None,
):
    """Send a message over XMPP."""

    class SendNotificationBot(slixmpp.ClientXMPP):
        """Service for sending Jabber (XMPP) messages."""

        def __init__(self):
            """Initialize the Jabber Bot."""
            super().__init__(sender, password)

            self.loop = hass.loop

            self.force_starttls = use_tls
            self.use_ipv6 = False
            self.add_event_handler("failed_auth", self.disconnect_on_login_fail)
            self.add_event_handler("session_start", self.start)

            if room:
                self.register_plugin("xep_0045")  # MUC
            if not verify_certificate:
                self.add_event_handler(
                    "ssl_invalid_cert", self.discard_ssl_invalid_cert
                )
            if data:
                # Init XEPs for image sending
                self.register_plugin("xep_0030")  # OOB dep
                self.register_plugin("xep_0066")  # Out of Band Data
                self.register_plugin("xep_0071")  # XHTML IM
                self.register_plugin("xep_0128")  # Service Discovery
                self.register_plugin("xep_0363")  # HTTP upload

            self.connect(force_starttls=self.force_starttls, use_ssl=False)

        async def start(self, event):
            """Start the communication and sends the message."""
            # Sending image and message independently from each other
            if data:
                await self.send_file(timeout=timeout)
            if message:
                self.send_text_message()

            self.disconnect()

        async def send_file(self, timeout=None):
            """Send file via XMPP.

            Send XMPP file message using OOB (XEP_0066) and
            HTTP Upload (XEP_0363)
            """
            if room:
                self.plugin["xep_0045"].join_muc(room, sender)

            try:
                # Uploading with XEP_0363
                _LOGGER.debug("Timeout set to %ss", timeout)
                url = await self.upload_file(timeout=timeout)

                _LOGGER.info("Upload success")
                for recipient in recipients:
                    if room:
                        _LOGGER.info("Sending file to %s", room)
                        message = self.Message(sto=room, stype="groupchat")
                    else:
                        _LOGGER.info("Sending file to %s", recipient)
                        message = self.Message(sto=recipient, stype="chat")
                    message["body"] = url
                    message["oob"]["url"] = url
                    try:
                        message.send()
                    except (IqError, IqTimeout, XMPPError) as ex:
                        _LOGGER.error("Could not send image message %s", ex)
                    if room:
                        break
            except (IqError, IqTimeout, XMPPError) as ex:
                _LOGGER.error("Upload error, could not send message %s", ex)
            except NotConnectedError as ex:
                _LOGGER.error("Connection error %s", ex)
            except FileTooBig as ex:
                _LOGGER.error("File too big for server, could not upload file %s", ex)
            except UploadServiceNotFound as ex:
                _LOGGER.error("UploadServiceNotFound, could not upload file %s", ex)
            except FileUploadError as ex:
                _LOGGER.error("FileUploadError, could not upload file %s", ex)
            except requests.exceptions.SSLError as ex:
                _LOGGER.error("Cannot establish SSL connection %s", ex)
            except requests.exceptions.ConnectionError as ex:
                _LOGGER.error("Cannot connect to server %s", ex)
            except (
                FileNotFoundError,
                PermissionError,
                IsADirectoryError,
                TimeoutError,
            ) as ex:
                _LOGGER.error("Error reading file %s", ex)
            except FutTimeoutError as ex:
                _LOGGER.error("The server did not respond in time, %s", ex)

        async def upload_file(self, timeout=None):
            """Upload file to Jabber server and return new URL.

            upload a file with Jabber XEP_0363 from a remote URL or a local
            file path and return a URL of that file.
            """
            if data.get(ATTR_URL_TEMPLATE):
                _LOGGER.debug("Got url template: %s", data[ATTR_URL_TEMPLATE])
                templ = template_helper.Template(data[ATTR_URL_TEMPLATE], hass)
                get_url = template_helper.render_complex(templ, None)
                url = await self.upload_file_from_url(get_url, timeout=timeout)
            elif data.get(ATTR_URL):
                url = await self.upload_file_from_url(data[ATTR_URL], timeout=timeout)
            elif data.get(ATTR_PATH_TEMPLATE):
                _LOGGER.debug("Got path template: %s", data[ATTR_PATH_TEMPLATE])
                templ = template_helper.Template(data[ATTR_PATH_TEMPLATE], hass)
                get_path = template_helper.render_complex(templ, None)
                url = await self.upload_file_from_path(get_path, timeout=timeout)
            elif data.get(ATTR_PATH):
                url = await self.upload_file_from_path(data[ATTR_PATH], timeout=timeout)
            else:
                url = None

            if url is None:
                _LOGGER.error("No path or URL found for file")
                raise FileUploadError("Could not upload file")

            return url

        async def upload_file_from_url(self, url, timeout=None):
            """Upload a file from a URL. Returns a URL.

            uploaded via XEP_0363 and HTTP and returns the resulting URL
            """
            _LOGGER.info("Getting file from %s", url)

            def get_url(url):
                """Return result for GET request to url."""
                return requests.get(
                    url, verify=data.get(ATTR_VERIFY, True), timeout=timeout
                )

            result = await hass.async_add_executor_job(get_url, url)

            if result.status_code >= HTTP_BAD_REQUEST:
                _LOGGER.error("Could not load file from %s", url)
                return None

            filesize = len(result.content)

            # we need a file extension, the upload server needs a
            # filename, if none is provided, through the path we guess
            # the extension
            # also setting random filename for privacy
            if data.get(ATTR_PATH):
                # using given path as base for new filename. Don't guess type
                filename = self.get_random_filename(data.get(ATTR_PATH))
            else:
                extension = (
                    mimetypes.guess_extension(result.headers["Content-Type"])
                    or ".unknown"
                )
                _LOGGER.debug("Got %s extension", extension)
                filename = self.get_random_filename(None, extension=extension)

            _LOGGER.info("Uploading file from URL, %s", filename)

            url = await self["xep_0363"].upload_file(
                filename,
                size=filesize,
                input_file=result.content,
                content_type=result.headers["Content-Type"],
                timeout=timeout,
            )

            return url

        async def upload_file_from_path(self, path, timeout=None):
            """Upload a file from a local file path via XEP_0363."""
            _LOGGER.info("Uploading file from path, %s", path)

            if not hass.config.is_allowed_path(path):
                raise PermissionError("Could not access file. Path not allowed")

            with open(path, "rb") as upfile:
                _LOGGER.debug("Reading file %s", path)
                input_file = upfile.read()
            filesize = len(input_file)
            _LOGGER.debug("Filesize is %s bytes", filesize)

            content_type = mimetypes.guess_type(path)[0]
            if content_type is None:
                content_type = DEFAULT_CONTENT_TYPE
            _LOGGER.debug("Content type is %s", content_type)

            # set random filename for privacy
            filename = self.get_random_filename(data.get(ATTR_PATH))
            _LOGGER.debug("Uploading file with random filename %s", filename)

            url = await self["xep_0363"].upload_file(
                filename,
                size=filesize,
                input_file=input_file,
                content_type=content_type,
                timeout=timeout,
            )

            return url

        def send_text_message(self):
            """Send a text only message to a room or a recipient."""
            try:
                if room:
                    _LOGGER.debug("Joining room %s", room)
                    self.plugin["xep_0045"].join_muc(room, sender)
                    self.send_message(mto=room, mbody=message, mtype="groupchat")
                else:
                    for recipient in recipients:
                        _LOGGER.debug("Sending message to %s", recipient)
                        self.send_message(mto=recipient, mbody=message, mtype="chat")
            except (IqError, IqTimeout, XMPPError) as ex:
                _LOGGER.error("Could not send text message %s", ex)
            except NotConnectedError as ex:
                _LOGGER.error("Connection error %s", ex)

        # pylint: disable=no-self-use
        def get_random_filename(self, filename, extension=None):
            """Return a random filename, leaving the extension intact."""
            if extension is None:
                path = pathlib.Path(filename)
                if path.suffix:
                    extension = "".join(path.suffixes)
                else:
                    extension = ".txt"
            return (
                "".join(random.choice(string.ascii_letters) for i in range(10))
                + extension
            )

        def disconnect_on_login_fail(self, event):
            """Disconnect from the server if credentials are invalid."""
            _LOGGER.warning("Login failed")
            self.disconnect()

        @staticmethod
        def discard_ssl_invalid_cert(event):
            """Do nothing if ssl certificate is invalid."""
            _LOGGER.info("Ignoring invalid SSL certificate as requested")

    SendNotificationBot()



"""Support for the Environment Canada weather service."""
from datetime import datetime, timedelta
import logging
import re

from env_canada import ECData
import voluptuous as vol

from homeassistant.components.sensor import PLATFORM_SCHEMA, SensorEntity
from homeassistant.const import (
    ATTR_ATTRIBUTION,
    ATTR_LOCATION,
    CONF_LATITUDE,
    CONF_LONGITUDE,
    TEMP_CELSIUS,
)
import homeassistant.helpers.config_validation as cv

_LOGGER = logging.getLogger(__name__)

SCAN_INTERVAL = timedelta(minutes=10)

ATTR_UPDATED = "updated"
ATTR_STATION = "station"
ATTR_TIME = "alert time"

CONF_ATTRIBUTION = "Data provided by Environment Canada"
CONF_STATION = "station"
CONF_LANGUAGE = "language"


def validate_station(station):
    """Check that the station ID is well-formed."""
    if station is None:
        return
    if not re.fullmatch(r"[A-Z]{2}/s0000\d{3}", station):
        raise vol.error.Invalid('Station ID must be of the form "XX/s0000###"')
    return station


PLATFORM_SCHEMA = PLATFORM_SCHEMA.extend(
    {
        vol.Required(CONF_LANGUAGE, default="english"): vol.In(["english", "french"]),
        vol.Optional(CONF_STATION): validate_station,
        vol.Inclusive(CONF_LATITUDE, "latlon"): cv.latitude,
        vol.Inclusive(CONF_LONGITUDE, "latlon"): cv.longitude,
    }
)


def setup_platform(hass, config, add_entities, discovery_info=None):
    """Set up the Environment Canada sensor."""

    if config.get(CONF_STATION):
        ec_data = ECData(
            station_id=config[CONF_STATION], language=config.get(CONF_LANGUAGE)
        )
    else:
        lat = config.get(CONF_LATITUDE, hass.config.latitude)
        lon = config.get(CONF_LONGITUDE, hass.config.longitude)
        ec_data = ECData(coordinates=(lat, lon), language=config.get(CONF_LANGUAGE))

    sensor_list = list(ec_data.conditions) + list(ec_data.alerts)
    add_entities([ECSensor(sensor_type, ec_data) for sensor_type in sensor_list], True)


class ECSensor(SensorEntity):
    """Implementation of an Environment Canada sensor."""

    def __init__(self, sensor_type, ec_data):
        """Initialize the sensor."""
        self.sensor_type = sensor_type
        self.ec_data = ec_data

        self._unique_id = None
        self._name = None
        self._state = None
        self._attr = None
        self._unit = None

    @property
    def unique_id(self) -> str:
        """Return the unique ID of the sensor."""
        return self._unique_id

    @property
    def name(self):
        """Return the name of the sensor."""
        return self._name

    @property
    def state(self):
        """Return the state of the sensor."""
        return self._state

    @property
    def extra_state_attributes(self):
        """Return the state attributes of the device."""
        return self._attr

    @property
    def unit_of_measurement(self):
        """Return the units of measurement."""
        return self._unit

    def update(self):
        """Update current conditions."""
        self.ec_data.update()
        self.ec_data.conditions.update(self.ec_data.alerts)

        conditions = self.ec_data.conditions
        metadata = self.ec_data.metadata
        sensor_data = conditions.get(self.sensor_type)

        self._unique_id = f"{metadata['location']}-{self.sensor_type}"
        self._attr = {}
        self._name = sensor_data.get("label")
        value = sensor_data.get("value")

        if isinstance(value, list):
            self._state = " | ".join([str(s.get("title")) for s in value])[:255]
            self._attr.update(
                {ATTR_TIME: " | ".join([str(s.get("date")) for s in value])}
            )
        elif self.sensor_type == "tendency":
            self._state = str(value).capitalize()
        elif value is not None and len(value) > 255:
            self._state = value[:255]
            _LOGGER.info("Value for %s truncated to 255 characters", self._unique_id)
        else:
            self._state = value

        if sensor_data.get("unit") == "C" or self.sensor_type in [
            "wind_chill",
            "humidex",
        ]:
            self._unit = TEMP_CELSIUS
        else:
            self._unit = sensor_data.get("unit")

        timestamp = metadata.get("timestamp")
        if timestamp:
            updated_utc = datetime.strptime(timestamp, "%Y%m%d%H%M%S").isoformat()
        else:
            updated_utc = None

        self._attr.update(
            {
                ATTR_ATTRIBUTION: CONF_ATTRIBUTION,
                ATTR_UPDATED: updated_utc,
                ATTR_LOCATION: metadata.get("location"),
                ATTR_STATION: metadata.get("station"),
            }
        )



"""Support for KNX/IP switches."""
from __future__ import annotations

from typing import Any, Callable, Iterable

from xknx.devices import Switch as XknxSwitch

from homeassistant.components.switch import SwitchEntity
from homeassistant.core import HomeAssistant
from homeassistant.helpers.entity import Entity
from homeassistant.helpers.typing import ConfigType, DiscoveryInfoType

from .const import DOMAIN
from .knx_entity import KnxEntity


async def async_setup_platform(
    hass: HomeAssistant,
    config: ConfigType,
    async_add_entities: Callable[[Iterable[Entity]], None],
    discovery_info: DiscoveryInfoType | None = None,
) -> None:
    """Set up switch(es) for KNX platform."""
    entities = []
    for device in hass.data[DOMAIN].xknx.devices:
        if isinstance(device, XknxSwitch):
            entities.append(KNXSwitch(device))
    async_add_entities(entities)


class KNXSwitch(KnxEntity, SwitchEntity):
    """Representation of a KNX switch."""

    def __init__(self, device: XknxSwitch) -> None:
        """Initialize of KNX switch."""
        self._device: XknxSwitch
        super().__init__(device)

    @property
    def is_on(self) -> bool:
        """Return true if device is on."""
        return bool(self._device.state)

    async def async_turn_on(self, **kwargs: Any) -> None:
        """Turn the device on."""
        await self._device.set_on()

    async def async_turn_off(self, **kwargs: Any) -> None:
        """Turn the device off."""
        await self._device.set_off()



"""Constants for Stream component."""
DOMAIN = "stream"

ATTR_ENDPOINTS = "endpoints"
ATTR_STREAMS = "streams"

OUTPUT_FORMATS = ["hls"]

SEGMENT_CONTAINER_FORMAT = "mp4"  # format for segments
RECORDER_CONTAINER_FORMAT = "mp4"  # format for recorder output
AUDIO_CODECS = {"aac", "mp3"}

FORMAT_CONTENT_TYPE = {"hls": "application/vnd.apple.mpegurl"}

OUTPUT_IDLE_TIMEOUT = 300  # Idle timeout due to inactivity

NUM_PLAYLIST_SEGMENTS = 3  # Number of segments to use in HLS playlist
MAX_SEGMENTS = 4  # Max number of segments to keep around
MIN_SEGMENT_DURATION = 1.5  # Each segment is at least this many seconds

PACKETS_TO_WAIT_FOR_AUDIO = 20  # Some streams have an audio stream with no audio
MAX_TIMESTAMP_GAP = 10000  # seconds - anything from 10 to 50000 is probably reasonable

MAX_MISSING_DTS = 6  # Number of packets missing DTS to allow
STREAM_TIMEOUT = 30  # Timeout for reading stream

STREAM_RESTART_INCREMENT = 10  # Increase wait_timeout by this amount each retry
STREAM_RESTART_RESET_TIME = 300  # Reset wait_timeout after this many seconds



"""Support for AIS dom MQTT sensors."""
import json
import logging
from typing import Optional

from homeassistant.components.mqtt import CONF_QOS, CONF_STATE_TOPIC, subscription
from homeassistant.const import CONF_NAME, CONF_UNIQUE_ID
from homeassistant.core import callback
from homeassistant.helpers.entity import Entity

_LOGGER = logging.getLogger(__name__)
G_RF_CODES = []


async def async_setup_platform(hass, config, async_add_entities, discovery_info=None):
    """Set up MQTT sensors through configuration.yaml."""
    config_entry = {
        CONF_NAME: "Ais Dom Mqtt RF Sensor",
        CONF_UNIQUE_ID: "ais_dom_mqtt_rf_sensor",
        CONF_QOS: 0,
        CONF_STATE_TOPIC: "tele/+/RESULT",
    }
    async_add_entities([MqttAisDomSensor(config_entry, "ais_dom_mqtt_sensor")])


class MqttAisDomSensor(Entity):
    """Representation of a sensor that can be updated using MQTT."""

    def __init__(self, config_entry, entity_ids):
        """Initialize the sensor."""
        self._config = config_entry
        self._unique_id = config_entry.get(CONF_UNIQUE_ID)
        self._state = None
        self._sub_state = None
        self._codes = []
        self._device_attributes = None

    async def async_added_to_hass(self):
        """Subscribe to MQTT events."""
        await self._subscribe_topics()

    async def _subscribe_topics(self):
        """(Re)Subscribe to topics."""

        @callback
        def message_received(msg):
            """Handle new MQTT messages."""
            try:
                payload = json.loads(msg.payload)
                if "RfRaw" in payload:
                    b1_code = payload.get("RfRaw")["Data"]
                    if b1_code == "AAA055":
                        # 'A0' means 'ACK'
                        return
                    topic = msg.topic
                    b0_code = get_b0_from_b1(b1_code)
                    code = {"topic": topic, "B1": b1_code, "B0": b0_code}
                    G_RF_CODES.append(code)
                    self._state = payload
                    self.async_write_ha_state()

            except Exception as e:
                _LOGGER.info("Error: " + str(e))

        self._sub_state = await subscription.async_subscribe_topics(
            self.hass,
            self._sub_state,
            {
                "state_topic": {
                    "topic": self._config[CONF_STATE_TOPIC],
                    "msg_callback": message_received,
                    "qos": self._config[CONF_QOS],
                }
            },
        )

    async def async_will_remove_from_hass(self):
        """Unsubscribe when removed."""
        self._sub_state = await subscription.async_unsubscribe_topics(
            self.hass, self._sub_state
        )

    @property
    def should_poll(self):
        """No polling needed."""
        return False

    @property
    def name(self):
        """Return the name of the sensor."""
        return self._config[CONF_NAME]

    @property
    def unit_of_measurement(self):
        """Return the unit this state is expressed in."""
        return ""

    @property
    def force_update(self):
        """Force update."""
        return False

    @property
    def state(self):
        """Return the state of the entity."""
        return self._state

    @property
    def device_state_attributes(self):
        """Return the state attributes."""
        return self._device_attributes

    @property
    def state_attributes(self):
        """Return the state attributes"""
        return {"codes": G_RF_CODES}

    @property
    def unique_id(self):
        """Return a unique ID."""
        return self._unique_id

    @property
    def icon(self):
        """Return the icon."""
        return "mdi:stop"

    @property
    def device_class(self) -> Optional[str]:
        """Return the device class of the sensor."""
        return None


# transfer b0 to b1 code
def get_b0_from_b1(b1):
    try:
        b1 = b1.replace(" ", "")
        iNbrOfBuckets = int(b1[4:6], 16)
        arrBuckets = []

        # Start packing
        sz_out_aux = " %0.2X " % iNbrOfBuckets
        sz_out_aux += "%0.2X " % int(8)

        for i in range(0, iNbrOfBuckets):
            sz_out_aux += b1[6 + i * 4 : 10 + i * 4] + " "
            arrBuckets.append(int(b1[6 + i * 4 : 10 + i * 4], 16))

        sz_out_aux += b1[10 + i * 4 : -2]

        szDataStr = sz_out_aux.replace(" ", "")
        sz_out_aux += " 55"
        iLength = int(len(szDataStr) / 2)
        sz_out_aux = "AA B0 " + "%0.2X" % iLength + sz_out_aux

        return sz_out_aux
    except Exception:
        _LOGGER.warning("Problem with b1 to b0 code transfer")
        return b1
